import hypothesis.strategies as st
from hypothesis import assume
from hypothesis import given

from skspatial.objects import Line
from skspatial.objects import Plane
from skspatial.objects import Points
from tests.property.constants import ATOL
from tests.property.strategies import arrays_fixed
from tests.property.strategies import lines
from tests.property.strategies import planes


@given(st.data())
def test_best_fit_line(data):

    n_points = data.draw(st.integers(min_value=2, max_value=5))
    dim = data.draw(st.integers(min_value=2, max_value=4))

    points = Points([data.draw(arrays_fixed(dim)) for _ in range(n_points)])
    assume(not points.are_concurrent(tol=ATOL))

    line = data.draw(lines(dim))
    line_fit = Line.best_fit(points)

    error_line = line.sum_squares(points)
    error_fit = line_fit.sum_squares(points)

    assert error_fit <= error_line + ATOL


@given(st.data())
def test_best_fit_plane(data):

    n_points = data.draw(st.integers(min_value=3, max_value=5))

    points = Points([data.draw(arrays_fixed(3)) for _ in range(n_points)])
    assume(not points.are_collinear(tol=ATOL))

    plane_fit = Plane.best_fit(points)

    # The best fit plane could have a higher dimension than the points
    # (e.g., 2D points have a 3D plane of best fit).
    # So, we convert the points dimension to that of the best fit plane.
    dim_fit = plane_fit.dimension
    points = points.set_dimension(dim_fit)

    plane = data.draw(planes(dim_fit))

    error_plane = plane.sum_squares(points)
    error_fit = plane_fit.sum_squares(points)

    assert error_fit <= error_plane + ATOL



"""Module for the Plane class."""
from __future__ import annotations

from typing import Optional
from typing import Tuple

import numpy as np
from mpl_toolkits.mplot3d import Axes3D

from skspatial.objects._base_line_plane import _BaseLinePlane
from skspatial.objects._mixins import _ToPointsMixin
from skspatial.objects.line import Line
from skspatial.objects.point import Point
from skspatial.objects.points import Points
from skspatial.objects.vector import Vector
from skspatial.typing import array_like


class Plane(_BaseLinePlane, _ToPointsMixin):
    """
    A plane in space.

    The plane is defined by a point and a normal vector.

    Parameters
    ----------
    point : array_like
        Point on the plane.
    direction : array_like
        Normal vector of the plane.
    kwargs : dict, optional
        Additional keywords passed to :meth:`Vector.is_zero`.
        This method is used to ensure that the normal vector is not the zero vector.

    Attributes
    ----------
    point : Point
        Point on the plane.
    normal : Vector
        Unit normal vector.
    vector : Vector
        Same as the normal.
    dimension : int
        Dimension of the plane.

    Raises
    ------
    ValueError
        If the point and vector have different dimensions.
        If the vector is all zeros.

    Examples
    --------
    >>> from skspatial.objects import Plane

    >>> plane = Plane(point=[0, 0, 0], normal=[0, 0, 5])

    >>> plane
    Plane(point=Point([0, 0, 0]), normal=Vector([0, 0, 5]))

    >>> plane.normal
    Vector([0, 0, 5])

    The normal can also be accessed with the ``vector`` attribute.

    >>> plane.vector
    Vector([0, 0, 5])

    The plane dimension is the dimension of the point and vector.

    >>> plane.dimension
    3

    >>> Plane([0, 0], [1, 0, 0])
    Traceback (most recent call last):
    ...
    ValueError: The point and vector must have the same dimension.

    >>> Plane([1, 1], [0, 0])
    Traceback (most recent call last):
    ...
    ValueError: The vector must not be the zero vector.

    """

    def __init__(self, point: array_like, normal: array_like):

        super().__init__(point, normal)
        self.normal = self.vector

    @classmethod
    def from_vectors(cls, point: array_like, vector_a: array_like, vector_b: array_like, **kwargs) -> Plane:
        """
        Instantiate a plane from a point and two vectors.

        The two vectors span the plane.

        Parameters
        ----------
        point : array_like
            Point on the plane.
        vector_a, vector_b : array_like
            Input vectors.
        kwargs : dict, optional
            Additional keywords passed to :meth:`Vector.is_parallel`.

        Returns
        -------
        Plane
            Plane containing input point and spanned by the two input vectors.

        Raises
        ------
        ValueError
            If the vectors are parallel.

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> Plane.from_vectors([0, 0], [1, 0], [0, 1])
        Plane(point=Point([0, 0, 0]), normal=Vector([0, 0, 1]))

        >>> Plane.from_vectors([0, 0], [1, 0], [2, 0])
        Traceback (most recent call last):
        ...
        ValueError: The vectors must not be parallel.

        """
        vector_a = Vector(vector_a)

        if vector_a.is_parallel(vector_b, **kwargs):
            raise ValueError("The vectors must not be parallel.")

        # The cross product returns a 3D vector.
        vector_normal = vector_a.cross(vector_b)

        # Convert the point to 3D so that it matches the vector dimension.
        point = Point(point).set_dimension(3)

        return cls(point, vector_normal)

    @classmethod
    def from_points(cls, point_a: array_like, point_b: array_like, point_c: array_like, **kwargs) -> Plane:
        """
        Instantiate a plane from three points.

        The three points lie on the plane.

        Parameters
        ----------
        point_a, point_b, point_c: array_like
            Three points defining the plane.
        kwargs: dict, optional
            Additional keywords passed to :meth:`Points.are_collinear`.

        Returns
        -------
        Plane
            Plane containing the three input points.

        Raises
        ------
        ValueError
            If the points are collinear.

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> Plane.from_points([0, 0], [1, 0], [3, 3])
        Plane(point=Point([0, 0, 0]), normal=Vector([0, 0, 3]))

        The order of the points affects the direction of the normal vector.

        >>> Plane.from_points([0, 0], [3, 3], [1, 0])
        Plane(point=Point([0, 0, 0]), normal=Vector([ 0,  0, -3]))

        >>> Plane.from_points([0, 0], [0, 1], [0, 3])
        Traceback (most recent call last):
        ...
        ValueError: The points must not be collinear.

        """
        if Points([point_a, point_b, point_c]).are_collinear(**kwargs):
            raise ValueError("The points must not be collinear.")

        vector_ab = Vector.from_points(point_a, point_b)
        vector_ac = Vector.from_points(point_a, point_c)

        return Plane.from_vectors(point_a, vector_ab, vector_ac)

    def cartesian(self) -> Tuple[np.number, np.number, np.number, np.number]:
        """
        Return the coefficients of the Cartesian equation of the plane.

        The equation has the form ax + by + cz + d = 0.

        Returns
        -------
        tuple
            Coefficients a, b, c, d.

        Raises
        ------
        ValueError
            If the plane dimension is higher than 3.

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> Plane(point=[1, 1], normal=[1, 0]).cartesian()
        (1, 0, 0, -1)

        >>> Plane(point=[1, 2, 0], normal=[0, 0, 1]).cartesian()
        (0, 0, 1, 0)

        >>> Plane(point=[1, 2, 8], normal=[0, 0, 5]).cartesian()
        (0, 0, 5, -40)

        >>> Plane(point=[4, 9, -1], normal=[10, 2, 4]).cartesian()
        (10, 2, 4, -54)

        >>> Plane([0, 0, 0, 0], [1, 0, 0, 0]).cartesian()
        Traceback (most recent call last):
        ...
        ValueError: The plane dimension must be <= 3.

        """
        if self.dimension > 3:
            raise ValueError("The plane dimension must be <= 3.")

        # The normal must be 3D to extract the coefficients.
        a, b, c = self.normal.set_dimension(3)

        d = -self.normal.dot(self.point)

        return a, b, c, d

    def project_point(self, point: array_like) -> Point:
        """
        Project a point onto the plane.

        Parameters
        ----------
        point : array_like
            Input point.

        Returns
        -------
        Point
            Projection of the point onto the plane.

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> plane = Plane(point=[0, 0, 0], normal=[0, 0, 2])

        >>> plane.project_point([10, 2, 5])
        Point([10.,  2.,  0.])

        >>> plane.project_point([5, 9, -3])
        Point([5., 9., 0.])

        """
        # Vector from the point in space to the point on the plane.
        vector_to_plane = Vector.from_points(point, self.point)

        # Perpendicular vector from the point in space to the plane.
        vector_projected = self.normal.project_vector(vector_to_plane)

        return Point(point) + vector_projected

    def project_vector(self, vector: array_like) -> Vector:
        """
        Project a vector onto the plane.

        Parameters
        ----------
        vector : array_like
            Input vector.

        Returns
        -------
        Vector
            Projection of the vector onto the plane.

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> plane = Plane([0, 4, 0], [0, 1, 1])

        >>> plane.project_vector([2, 4, 8])
        Vector([ 2., -2.,  2.])

        """
        point_in_space = self.point + vector
        point_on_plane = self.project_point(point_in_space)

        return Vector.from_points(self.point, point_on_plane)

    def project_line(self, line: Line, **kwargs: float) -> Line:
        """
        Project a line onto the plane.

        This method can also handle the case where the line is parallel to the plane.

        Parameters
        ----------
        line : Line
            Input line.
        kwargs : dict, optional
            Additional keywords passed to :meth:`Vector.is_perpendicular`,
            which is used to check if the line is parallel to the plane
            (i.e., the line direction is perpendicular to the plane normal).

        Returns
        -------
        Line
            Projection of the line onto the plane.

        Examples
        --------
        >>> from skspatial.objects import Line, Plane

        >>> plane = Plane([0, 0, 0], [0, 0, 1])
        >>> line = Line([0, 0, 0], [1, 1, 1])

        >>> plane.project_line(line)
        Line(point=Point([0., 0., 0.]), direction=Vector([1., 1., 0.]))

        The line is parallel to the plane.

        >>> line = Line([0, 0, 5], [1, 0, 0])

        >>> plane.project_line(line)
        Line(point=Point([0., 0., 0.]), direction=Vector([1, 0, 0]))

        """
        if self.normal.is_parallel(line.vector, **kwargs):
            raise ValueError("The line and plane must not be perpendicular.")

        point_projected = self.project_point(line.point)

        if self.normal.is_perpendicular(line.vector, **kwargs):
            return Line(point_projected, line.vector)

        vector_projected = self.project_vector(line.vector)

        return Line(point_projected, vector_projected)

    def distance_point_signed(self, point: array_like) -> np.float64:
        """
        Return the signed distance from a point to the plane.

        Parameters
        ----------
        point : array_like
            Input point.

        Returns
        -------
        np.float64
            Signed distance from the point to the plane.

        References
        ----------
        http://mathworld.wolfram.com/Point-PlaneDistance.html

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> plane = Plane([0, 0, 0], [0, 0, 1])

        >>> plane.distance_point_signed([5, 2, 0])
        0.0

        >>> plane.distance_point_signed([5, 2, 1])
        1.0

        >>> plane.distance_point([5, 2, -4])
        4.0
        >>> plane.distance_point_signed([5, 2, -4])
        -4.0

        """
        vector_to_point = Vector.from_points(self.point, point)

        return self.normal.scalar_projection(vector_to_point)

    def distance_point(self, point: array_like) -> np.float64:
        """
        Return the distance from a point to the plane.

        Parameters
        ----------
        point : array_like
            Input point.

        Returns
        -------
        np.float64
            Distance from the point to the plane.

        References
        ----------
        http://mathworld.wolfram.com/Point-PlaneDistance.html

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> plane = Plane([0, 0, 0], [0, 0, 1])

        >>> plane.distance_point([5, 2, 0])
        0.0

        >>> plane.distance_point([5, 2, 1])
        1.0

        >>> plane.distance_point([5, 2, -4])
        4.0

        """
        return abs(self.distance_point_signed(point))

    def side_point(self, point: array_like) -> int:
        """
        Find the side of the plane where a point lies.

        Parameters
        ----------
        point : array_like
            Input point.

        Returns
        -------
        int
            -1 if the point is behind the plane.
            0 if the point is on the plane.
            1 if the point is in front of the plane.

        Examples
        --------
        >>> from skspatial.objects import Plane
        >>> plane = Plane([0, 0, 0], [0, 0, 1])

        The point is in on the plane.

        >>> plane.side_point([2, 5, 0])
        0

        The point is in front of the plane.

        >>> plane.side_point([1, -5, 6])
        1

        The point is behind the plane.

        >>> plane.side_point([5, 8, -4])
        -1

        Higher dimensions are supported.

        >>> plane = Plane([0, 0, 0, 0], [0, 1, 0, 1])
        >>> plane.side_point([0, -10, 4, 1])
        -1

        """
        return int(np.sign(self.distance_point_signed(point)))

    def intersect_line(self, line: Line, **kwargs) -> Point:
        """
        Intersect the plane with a line.

        The line and plane must not be parallel.

        Parameters
        ----------
        line : Line
            Input line.
        kwargs : dict, optional
            Additional keywords passed to :meth:`Vector.is_perpendicular`.

        Returns
        -------
        Point
            The point of intersection.

        Raises
        ------
        ValueError
            If the line and plane are parallel.

        References
        ----------
        http://geomalgorithms.com/a05-_intersect-1.html

        Examples
        --------
        >>> from skspatial.objects import Line, Plane

        >>> line = Line([0, 0, 0], [0, 0, 1])
        >>> plane = Plane([0, 0, 0], [0, 0, 1])

        >>> plane.intersect_line(line)
        Point([0., 0., 0.])

        >>> plane = Plane([2, -53, -7], [0, 0, 1])
        >>> plane.intersect_line(line)
        Point([ 0.,  0., -7.])

        >>> line = Line([0, 1, 0], [1, 0, 0])
        >>> plane.intersect_line(line)
        Traceback (most recent call last):
        ...
        ValueError: The line and plane must not be parallel.

        """
        if self.normal.is_perpendicular(line.direction, **kwargs):
            raise ValueError("The line and plane must not be parallel.")

        vector_plane_line = Vector.from_points(self.point, line.point)

        num = -self.normal.dot(vector_plane_line)
        denom = self.normal.dot(line.direction)

        # Vector along the line to the intersection point.
        vector_line_scaled = num / denom * line.direction

        return line.point + vector_line_scaled

    def intersect_plane(self, other: Plane, **kwargs) -> Line:
        """
        Intersect the plane with another.

        The planes must not be parallel.

        Parameters
        ----------
        other : Plane
            Other plane.
        kwargs : dict, optional
            Additional keywords passed to :meth:`Vector.is_parallel`.

        Returns
        -------
        Line
            The line of intersection.

        Raises
        ------
        ValueError
            If the planes are parallel.

        References
        ----------
        http://tbirdal.blogspot.com/2016/10/a-better-approach-to-plane-intersection.html

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> plane_a = Plane([0, 0, 0], [0, 0, 1])
        >>> plane_b = Plane([0, 0, 0], [1, 0, 0])

        >>> plane_a.intersect_plane(plane_b)
        Line(point=Point([0., 0., 0.]), direction=Vector([0, 1, 0]))

        >>> plane_b = Plane([5, 16, -94], [1, 0, 0])
        >>> plane_a.intersect_plane(plane_b)
        Line(point=Point([5., 0., 0.]), direction=Vector([0, 1, 0]))

        >>> plane_b = Plane([0, 0, 1], [1, 0, 1])
        >>> plane_a.intersect_plane(plane_b)
        Line(point=Point([1., 0., 0.]), direction=Vector([0, 1, 0]))

        >>> plane_b = Plane([0, 0, 5], [0, 0, -8])
        >>> plane_a.intersect_plane(plane_b)
        Traceback (most recent call last):
        ...
        ValueError: The planes must not be parallel.

        """
        if self.normal.is_parallel(other.normal, **kwargs):
            raise ValueError("The planes must not be parallel.")

        array_normals_stacked = np.vstack((self.normal, other.normal))

        # Construct a matrix for a linear system.
        array_00 = 2 * np.eye(3)
        array_01 = array_normals_stacked.T
        array_10 = array_normals_stacked
        array_11 = np.zeros((2, 2))
        matrix = np.block([[array_00, array_01], [array_10, array_11]])

        dot_a = np.dot(self.point, self.normal)
        dot_b = np.dot(other.point, other.normal)
        array_y = np.array([0, 0, 0, dot_a, dot_b])

        # Solve the linear system.
        solution = np.linalg.solve(matrix, array_y)

        point_line = Point(solution[:3])
        direction_line = self.normal.cross(other.normal)

        return Line(point_line, direction_line)

    @classmethod
    def best_fit(cls, points: array_like, tol: Optional[float] = None, **kwargs) -> Plane:
        """
        Return the plane of best fit for a set of 3D points.

        Parameters
        ----------
        points : array_like
             Input 3D points.
        tol : float | None, optional
            Keyword passed to :meth:`Points.are_collinear` (default None).
        kwargs : dict, optional
            Additional keywords passed to :func:`numpy.linalg.svd`

        Returns
        -------
        Plane
            The plane of best fit.

        Raises
        ------
        ValueError
            If the points are collinear or are not 3D.

        References
        ----------
        Using SVD for some fitting problems
        Inge Söderkvist
        Algorithm 3.1
        https://www.ltu.se/cms_fs/1.51590!/svd-fitting.pdf

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> points = [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]]
        >>> plane = Plane.best_fit(points)

        The point on the plane is the centroid of the points.

        >>> plane.point
        Point([0.25, 0.25, 0.25])

        The plane normal is a unit vector.

        >>> plane.normal.round(3)
        Vector([-0.577, -0.577, -0.577])

        >>> points = [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0]]

        >>> Plane.best_fit(points)
        Plane(point=Point([0.5, 0.5, 0. ]), normal=Vector([0., 0., 1.]))

        >>> Plane.best_fit(points, full_matrices=False)
        Plane(point=Point([0.5, 0.5, 0. ]), normal=Vector([0., 0., 1.]))

        """
        points = Points(points)

        if points.dimension != 3:
            raise ValueError("The points must be 3D.")

        if points.are_collinear(tol=tol):
            raise ValueError("The points must not be collinear.")

        points_centered, centroid = points.mean_center(return_centroid=True)

        u, _, _ = np.linalg.svd(points_centered.T, **kwargs)
        normal = Vector(u[:, 2])

        return cls(centroid, normal)

    def to_mesh(
        self,
        lims_x: array_like = (-1, 1),
        lims_y: array_like = (-1, 1),
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Return coordinate matrices for the 3D surface of the plane.

        Parameters
        ----------
        lims_x, lims_y : (2,) tuple
            x and y limits of the plane.
            Tuple of form (min, max). The default is (-1, 1).
            The point on the plane is used as the origin.

        Returns
        -------
        X, Y, Z: ndarray
            Coordinate matrices.

        Examples
        --------
        >>> from skspatial.objects import Plane

        >>> X, Y, Z = Plane([0, 0, 0], [0, 0, 1]).to_mesh()

        >>> X
        array([[-1,  1],
               [-1,  1]])

        >>> Y
        array([[-1, -1],
               [ 1,  1]])

        >>> Z
        array([[0., 0.],
               [0., 0.]])

        """
        a, b, c, d = self.cartesian()
        x_center, y_center = self.point[:2]

        values_x = x_center + lims_x
        values_y = y_center + lims_y

        X, Y = np.meshgrid(values_x, values_y)

        if c != 0:
            Z = -(a * X + b * Y + d) / c

        elif b != 0:
            Z = -(a * X + c * Y + d) / b
            X, Y, Z = X, Z, Y

        else:
            Z = -(b * X + c * Y + d) / a
            X, Y, Z = Z, X, Y

        return X, Y, Z

    def plot_3d(self, ax_3d: Axes3D, lims_x: array_like = (-1, 1), lims_y: array_like = (-1, 1), **kwargs) -> None:
        """
        Plot a 3D plane.

        Parameters
        ----------
        ax_3d : Axes3D
            Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
        lims_x, lims_y : (2,) tuple
            x and y limits of the plane.
            Tuple of form (min, max). The default is (-1, 1).
            The point on the plane is used as the origin.
        kwargs : dict, optional
            Additional keywords passed to :meth:`~mpl_toolkits.mplot3d.axes3d.Axes3D.plot_surface`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from mpl_toolkits.mplot3d import Axes3D

            >>> from skspatial.objects import Plane

            >>> fig = plt.figure()
            >>> ax = fig.add_subplot(111, projection='3d')

            >>> plane = Plane([5, 3, 1], [1, 0, 1])

            >>> plane.plot_3d(ax, alpha=0.2)
            >>> plane.point.plot_3d(ax, s=100)

        """
        X, Y, Z = self.to_mesh(lims_x, lims_y)

        ax_3d.plot_surface(X, Y, Z, **kwargs)



"""Module for the Line class."""
from __future__ import annotations

from typing import Optional

import numpy as np
from matplotlib.axes import Axes
from mpl_toolkits.mplot3d import Axes3D

from skspatial.objects._base_line_plane import _BaseLinePlane
from skspatial.objects.point import Point
from skspatial.objects.points import Points
from skspatial.objects.vector import Vector
from skspatial.plotting import _connect_points_2d
from skspatial.plotting import _connect_points_3d
from skspatial.transformation import transform_coordinates
from skspatial.typing import array_like


class Line(_BaseLinePlane):
    """
    A line in space.

    The line is defined by a point and a direction vector.

    Parameters
    ----------
    point : array_like
        Point on the line.
    direction : array_like
        Direction vector of the line.
    kwargs : dict, optional
        Additional keywords passed to :meth:`Vector.is_zero`.
        This method is used to ensure that the direction vector is not the zero vector.

    Attributes
    ----------
    point : Point
        Point on the line.
    direction : Vector
        Unit direction vector.
    vector : Vector
        Same as the direction.
    dimension : int
        Dimension of the line.

    Raises
    ------
    ValueError
        If the point and vector have different dimensions.
        If the vector is all zeros.

    Examples
    --------
    >>> from skspatial.objects import Line

    >>> line = Line(point=[0, 0], direction=[3, 0])

    >>> line
    Line(point=Point([0, 0]), direction=Vector([3, 0]))

    >>> line.direction
    Vector([3, 0])

    The direction can also be accessed with the ``vector`` attribute.

    >>> line.vector
    Vector([3, 0])

    The line dimension is the dimension of the point and vector.

    >>> line.dimension
    2

    >>> Line([0, 0], [1, 0, 0])
    Traceback (most recent call last):
    ...
    ValueError: The point and vector must have the same dimension.

    >>> Line([1, 1], [0, 0])
    Traceback (most recent call last):
    ...
    ValueError: The vector must not be the zero vector.

    """

    def __init__(self, point: array_like, direction: array_like):

        super().__init__(point, direction)

        self.direction = self.vector

    @classmethod
    def from_points(cls, point_a: array_like, point_b: array_like) -> Line:
        """
        Instantiate a line from two points.

        Parameters
        ----------
        point_a, point_b : array_like
            Two points defining the line.

        Returns
        -------
        Line
            Line containing the two input points.

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> Line.from_points([0, 0], [1, 0])
        Line(point=Point([0, 0]), direction=Vector([1, 0]))

        The order of the points affects the line point and direction vector.

        >>> Line.from_points([1, 0], [0, 0])
        Line(point=Point([1, 0]), direction=Vector([-1,  0]))

        """
        vector_ab = Vector.from_points(point_a, point_b)

        return cls(point_a, vector_ab)

    @classmethod
    def from_slope(cls, slope: float, y_intercept: float) -> Line:
        r"""
        Instantiate a 2D line from a slope and Y-intercept.

        A 2D line can be represented by the equation

        .. math:: y = mx + b

        where :math:`m` is the slope and :math:`p` is the Y-intercept.

        Parameters
        ----------
        slope : {int, float}
            Slope of the 2D line.
        y_intercept : {int, float}
            Y coordinate of the point where the line intersects the Y axis.

        Returns
        -------
        Line
            A 2D Line object.

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> Line.from_slope(2, 0)
        Line(point=Point([0, 0]), direction=Vector([1, 2]))

        >>> Line.from_slope(-3, 5)
        Line(point=Point([0, 5]), direction=Vector([ 1, -3]))

        >>> line_a = Line.from_slope(1, 0)
        >>> line_b = Line.from_slope(0, 5)

        >>> line_a.intersect_line(line_b)
        Point([5., 5.])

        """
        point = [0, y_intercept]
        direction = [1, slope]

        return cls(point, direction)

    def is_coplanar(self, other: Line, **kwargs: float) -> bool:
        """
        Check if the line is coplanar with another.

        Parameters
        ----------
        other : Line
            Other line.
        kwargs : dict, optional
            Additional keywords passed to :func:`numpy.linalg.matrix_rank`

        Returns
        -------
        bool
            True if the line is coplanar; false otherwise.

        Raises
        ------
        TypeError
            If the input is not a line.

        References
        ----------
        http://mathworld.wolfram.com/Coplanar.html

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> line_a = Line(point=[0, 0, 0], direction=[1, 0, 0])
        >>> line_b = Line([-5, 3, 0], [7, 1, 0])
        >>> line_c = Line([0, 0, 0], [0, 0, 1])

        >>> line_a.is_coplanar(line_b)
        True

        >>> line_a.is_coplanar(line_c)
        True

        >>> line_b.is_coplanar(line_c)
        False

        The input must be another line.

        >>> from skspatial.objects import Plane

        >>> line_a.is_coplanar(Plane(line_a.point, line_a.vector))
        Traceback (most recent call last):
        ...
        TypeError: The input must also be a line.

        """
        if not isinstance(other, type(self)):
            raise TypeError("The input must also be a line.")

        point_1 = self.point
        point_2 = self.to_point()
        point_3 = other.point
        point_4 = other.to_point()

        points = Points([point_1, point_2, point_3, point_4])

        return points.are_coplanar(**kwargs)

    def to_point(self, t: float = 1) -> Point:
        r"""
        Return a point along the line using a parameter `t`.

        Parameters
        ----------
        t : {int, float}
            Parameter that defines the new point along the line.

        Returns
        -------
        Point
            New point along the line.

        Notes
        -----
        The new point :math:`p` is computed as:

        .. math:: p = \mathtt{line.point} + t \cdot \mathtt{line.direction}

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> line = Line(point=[0, 0], direction=[2, 0])

        >>> line.to_point()
        Point([2, 0])

        >>> line.to_point(t=2)
        Point([4, 0])

        """
        vector_along_line = t * self.direction

        return self.point + vector_along_line

    def project_point(self, point: array_like) -> Point:
        """
        Project a point onto the line.

        Parameters
        ----------
        point : array_like
            Input point.

        Returns
        -------
        Point
            Projection of the point onto the line.

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> Line(point=[0, 0], direction=[8, 0]).project_point([5, 5])
        Point([5., 0.])

        >>> Line(point=[0, 0, 0], direction=[1, 1, 0]).project_point([5, 5, 3])
        Point([5., 5., 0.])

        """
        # Vector from the point on the line to the point in space.
        vector_to_point = Vector.from_points(self.point, point)

        # Project the vector onto the line.
        vector_projected = self.direction.project_vector(vector_to_point)

        # Add the projected vector to the point on the line.
        return self.point + vector_projected

    def project_vector(self, vector: array_like) -> Vector:
        """
        Project a vector onto the line.

        Parameters
        ----------
        vector : array_like
            Input vector.

        Returns
        -------
        Vector
            Projection of the vector onto the line.

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> line = Line([-1, 5, 3], [3, 4, 5])

        >>> line.project_vector([1, 1, 1])
        Vector([0.72, 0.96, 1.2 ])

        """
        return self.direction.project_vector(vector)

    def side_point(self, point: array_like) -> int:
        """
        Find the side of the line where a point lies.

        The line and point must be 2D.

        Parameters
        ----------
        point : array_like
            Input point.

        Returns
        -------
        int
            -1 if the point is left of the line.
            0 if the point is on the line.
            1 if the point is right of the line.

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> line = Line([0, 0], [1, 1])

        The point is on the line.

        >>> line.side_point([2, 2])
        0

        The point is to the right of the line.

        >>> line.side_point([5, 3])
        1

        The point is to the left of the line.

        >>> line.side_point([5, 10])
        -1

        """
        vector_to_point = Vector.from_points(self.point, point)

        return self.direction.side_vector(vector_to_point)

    def distance_point(self, point: array_like) -> np.float64:
        """
        Return the distance from a point to the line.

        This is the distance from the point to its projection on the line.

        Parameters
        ----------
        point : array_like
            Input point.

        Returns
        -------
        np.float64
            Distance from the point to the line.

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> line = Line([0, 0], [1, 0])
        >>> line.distance_point([0, 0])
        0.0
        >>> line.distance_point([5, 0])
        0.0
        >>> line.distance_point([5, -5])
        5.0

        >>> line = Line([5, 2, -3], [3, 8, 2])
        >>> line.distance_point([5, -5, 3]).round(3)
        7.737

        """
        point_projected = self.project_point(point)

        return point_projected.distance_point(point)

    def distance_line(self, other: Line) -> np.float64:
        """
        Return the shortest distance from the line to another.

        Parameters
        ----------
        other : Line
            Other line.

        Returns
        -------
        np.float64
            Distance between the lines.

        References
        ----------
        http://mathworld.wolfram.com/Line-LineDistance.html

        Examples
        --------
        There are three cases:

        1. The lines intersect (i.e., they are coplanar and not parallel).

        >>> from skspatial.objects import Line
        >>> line_a = Line([1, 2], [4, 3])
        >>> line_b = Line([-4, 1], [7, 23])

        >>> line_a.distance_line(line_b)
        0.0

        2. The lines are parallel.

        >>> line_a = Line([0, 0], [1, 0])
        >>> line_b = Line([0, 5], [-1, 0])

        >>> line_a.distance_line(line_b)
        5.0

        3. The lines are skew.

        >>> line_a = Line([0, 0, 0], [1, 0, 1])
        >>> line_b = Line([1, 0, 0], [1, 1, 1])

        >>> line_a.distance_line(line_b).round(3)
        0.707

        """
        if self.direction.is_parallel(other.direction):
            # The lines are parallel.
            # The distance between the lines is the distance from line point B to line A.
            distance = self.distance_point(other.point)

        elif self.is_coplanar(other):
            # The lines must intersect, since they are coplanar and not parallel.
            distance = np.float64(0)

        else:
            # The lines are skew.
            vector_ab = Vector.from_points(self.point, other.point)
            vector_perpendicular = self.direction.cross(other.direction)

            distance = abs(vector_ab.dot(vector_perpendicular)) / vector_perpendicular.norm()

        return distance

    def intersect_line(self, other: Line, **kwargs) -> Point:
        """
        Intersect the line with another.

        The lines must be coplanar and not parallel.

        Parameters
        ----------
        other : Line
            Other line.
        kwargs : dict, optional
            Additional keywords passed to :meth:`Vector.is_parallel`.

        Returns
        -------
        Point
            The point at the intersection.

        Raises
        ------
        ValueError
            If the lines don't have the same dimension.
            If the line dimension is greater than three.
            If the lines are parallel.
            If the lines are not coplanar.

        References
        ----------
        http://mathworld.wolfram.com/Line-LineIntersection.html

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> line_a = Line([0, 0], [1, 0])
        >>> line_b = Line([5, 5], [0, 1])
        >>> line_a.intersect_line(line_b)
        Point([5., 0.])

        >>> line_a = Line([0, 0, 0], [1, 1, 1])
        >>> line_b = Line([5, 5, 0], [0, 0, -8])
        >>> line_a.intersect_line(line_b)
        Point([5., 5., 5.])

        >>> line_a = Line([0, 0, 0], [1, 0, 0])
        >>> line_b = Line([0, 0], [1, 1])
        >>> line_a.intersect_line(line_b)
        Traceback (most recent call last):
        ...
        ValueError: The lines must have the same dimension.

        >>> line_a = Line(4 * [0], [1, 0, 0, 0])
        >>> line_b = Line(4 * [0], [0, 0, 0, 1])
        >>> line_a.intersect_line(line_b)
        Traceback (most recent call last):
        ...
        ValueError: The line dimension cannot be greater than 3.

        >>> line_a = Line([0, 0], [0, 1])
        >>> line_b = Line([0, 1], [0, 1])

        >>> line_a = Line([0, 0], [1, 0])
        >>> line_b = Line([0, 1], [2, 0])
        >>> line_a.intersect_line(line_b)
        Traceback (most recent call last):
        ...
        ValueError: The lines must not be parallel.

        >>> line_a = Line([1, 2, 3], [-4, 1, 1])
        >>> line_b = Line([4, 5, 6], [3, 1, 5])
        >>> line_a.intersect_line(line_b)
        Traceback (most recent call last):
        ...
        ValueError: The lines must be coplanar.

        """
        if self.dimension != other.dimension:
            raise ValueError("The lines must have the same dimension.")

        if self.dimension > 3 or other.dimension > 3:
            raise ValueError("The line dimension cannot be greater than 3.")

        if self.direction.is_parallel(other.direction, **kwargs):
            raise ValueError("The lines must not be parallel.")

        if not self.is_coplanar(other):
            raise ValueError("The lines must be coplanar.")

        # Vector from line A to line B.
        vector_ab = Vector.from_points(self.point, other.point)

        # Vector perpendicular to both lines.
        vector_perpendicular = self.direction.cross(other.direction)

        num = vector_ab.cross(other.direction).dot(vector_perpendicular)
        denom = vector_perpendicular.norm() ** 2

        # Vector along line A to the intersection point.
        vector_a_scaled = num / denom * self.direction

        return self.point + vector_a_scaled

    @classmethod
    def best_fit(cls, points: array_like, tol: Optional[float] = None, **kwargs) -> Line:
        """
        Return the line of best fit for a set of points.

        Parameters
        ----------
        points : array_like
             Input points.
        tol : float | None, optional
            Keyword passed to :meth:`Points.are_collinear` (default None).
        kwargs : dict, optional
            Additional keywords passed to :func:`numpy.linalg.svd`

        Returns
        -------
        Line
            The line of best fit.

        Raises
        ------
        ValueError
            If the points are concurrent.

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> points = [[0, 0], [1, 2], [2, 1], [2, 3], [3, 2]]
        >>> line = Line.best_fit(points)

        The point on the line is the centroid of the points.

        >>> line.point
        Point([1.6, 1.6])

        The line direction is a unit vector.

        >>> line.direction.round(3)
        Vector([0.707, 0.707])

        """
        points_spatial = Points(points)

        if points_spatial.are_concurrent(tol=tol):
            raise ValueError("The points must not be concurrent.")

        points_centered, centroid = points_spatial.mean_center(return_centroid=True)

        _, _, vh = np.linalg.svd(points_centered, **kwargs)
        direction = vh[0, :]

        return cls(centroid, direction)

    def transform_points(self, points: array_like) -> np.ndarray:
        """
        Transform points to a one-dimensional coordinate system defined by the line.

        The point on the line acts as the origin of the coordinate system.

        This is analogous is projecting the points onto the line,
        then computing the signed distance from the line point to the projections.

        Parameters
        ----------
        points : (N, D) array_like
            Array of N points with dimension D.

        Returns
        -------
        ndarray
            (N,) array of N coordinates.

        Examples
        --------
        >>> from skspatial.objects import Line

        >>> points = [[-1, 1], [0, 1], [1, 1], [2, 1]]

        >>> Line([0, 0], [1, 0]).transform_points(points)
        array([-1.,  0.,  1.,  2.])

        The point on the line acts as the origin of the coordinates.

        >>> Line([1, 0], [1, 0]).transform_points(points)
        array([-2., -1.,  0.,  1.])

        The sign of the coordinates depends on the direction of the line.

        >>> Line([0, 0], [-1, 0]).transform_points(points)
        array([ 1.,  0., -1., -2.])

        The magnitude of the direction vector is irrelevant.

        >>> Line([0, 0], [5, 0]).transform_points(points)
        array([-1.,  0.,  1.,  2.])

        """
        # Basis vector of the subspace (the line).
        vectors_basis = [self.direction.unit()]

        column = transform_coordinates(points, self.point.to_array(), vectors_basis)

        return column.flatten()

    def plot_2d(self, ax_2d: Axes, t_1: float = 0, t_2: float = 1, **kwargs) -> None:
        """
        Plot a 2D line.

        The line is plotted by connecting two 2D points.

        Parameters
        ----------
        ax_2d : Axes
            Instance of :class:`~matplotlib.axes.Axes`.
        t_1, t_2 : {int, float}
            Parameters to determine points 1 and 2 along the line.
            These are passed to :meth:`Line.to_point`.
            Defaults are 0 and 1.
        kwargs : dict, optional
            Additional keywords passed to :meth:`~matplotlib.axes.Axes.plot`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from skspatial.objects import Line

            >>> _, ax = plt.subplots()

            >>> line = Line([1, 2], [3, 4])

            >>> line.plot_2d(ax, t_1=-2, t_2=3, c='k')
            >>> line.point.plot_2d(ax, c='r', s=100, zorder=3)
            >>> grid = ax.grid()

        """
        point_1 = self.to_point(t_1)
        point_2 = self.to_point(t_2)

        _connect_points_2d(ax_2d, point_1, point_2, **kwargs)

    def plot_3d(self, ax_3d: Axes3D, t_1: float = 0, t_2: float = 1, **kwargs) -> None:
        """
        Plot a 3D line.

        The line is plotted by connecting two 3D points.

        Parameters
        ----------
        ax_3d : Axes3D
            Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
        t_1, t_2 : {int, float}
            Parameters to determine points 1 and 2 along the line.
            These are passed to :meth:`Line.to_point`.
            Defaults are 0 and 1.
        kwargs : dict, optional
            Additional keywords passed to :meth:`~mpl_toolkits.mplot3d.axes3d.Axes3D.plot`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from mpl_toolkits.mplot3d import Axes3D

            >>> from skspatial.objects import Line

            >>> fig = plt.figure()
            >>> ax = fig.add_subplot(111, projection='3d')

            >>> line = Line([1, 2, 3], [0, 1, 1])

            >>> line.plot_3d(ax, c='k')
            >>> line.point.plot_3d(ax, s=100)

        """
        point_1 = self.to_point(t_1)
        point_2 = self.to_point(t_2)

        _connect_points_3d(ax_3d, point_1, point_2, **kwargs)



"""Constants used for property tests."""
# Minimum absolute value of floats.
FLOAT_MIN = 1e-3

# Absolute tolerance for math.isclose.
ATOL = 1e-4

# Min and max dimensions of arrays.
DIM_MIN, DIM_MAX = 2, 10



"""Mixin classes."""
from typing import Callable
from typing import Tuple

import numpy as np

from skspatial._functions import _mesh_to_points
from skspatial.objects.points import Points


class _ToPointsMixin:

    to_mesh: Callable[..., Tuple[np.ndarray, np.ndarray, np.ndarray]]

    def to_points(self, **kwargs) -> Points:
        """
        Return points on the surface of the object.

        Parameters
        ----------
        kwargs: dict, optional
            Additional keywords passed to the `to_mesh` method of the class.

        Returns
        -------
        Points
            Points on the surface of the object.

        Examples
        --------
        >>> from skspatial.objects import Sphere

        >>> sphere = Sphere([0, 0, 0], 1)

        >>> sphere.to_points(n_angles=3).round().unique()
        Points([[ 0., -1.,  0.],
                [ 0.,  0., -1.],
                [ 0.,  0.,  1.],
                [ 0.,  1.,  0.]])

        >>> sphere.to_points(n_angles=4).round(3).unique()
        Points([[-0.75 , -0.433, -0.5  ],
                [-0.75 , -0.433,  0.5  ],
                [ 0.   ,  0.   , -1.   ],
                [ 0.   ,  0.   ,  1.   ],
                [ 0.   ,  0.866, -0.5  ],
                [ 0.   ,  0.866,  0.5  ],
                [ 0.75 , -0.433, -0.5  ],
                [ 0.75 , -0.433,  0.5  ]])

        """
        X, Y, Z = self.to_mesh(**kwargs)
        points = _mesh_to_points(X, Y, Z)

        return Points(points)



import math

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from skspatial.objects import Vector


@pytest.mark.parametrize(
    ("array_a", "array_b", "vector_expected"),
    [
        ([0, 0], [1, 0], Vector([1, 0])),
        ([1, 0], [1, 0], Vector([0, 0])),
        ([1, 0], [2, 0], Vector([1, 0])),
        ([8, 3, -5], [3, 7, 1], Vector([-5, 4, 6])),
        ([5, 7, 8, 9], [2, 5, 3, -4], Vector([-3, -2, -5, -13])),
    ],
)
def test_from_points(array_a, array_b, vector_expected):

    assert_array_equal(Vector.from_points(array_a, array_b), vector_expected)


@pytest.mark.parametrize(
    ("array", "array_unit_expected"),
    [
        ([1, 0], [1, 0]),
        ([2, 0], [1, 0]),
        ([-1, 0], [-1, 0]),
        ([0, 0, 5], [0, 0, 1]),
        ([1, 1], [math.sqrt(2) / 2, math.sqrt(2) / 2]),
        ([1, 1, 1], [math.sqrt(3) / 3, math.sqrt(3) / 3, math.sqrt(3) / 3]),
        ([2, 0, 0, 0], [1, 0, 0, 0]),
        ([3, 3, 0, 0], [math.sqrt(2) / 2, math.sqrt(2) / 2, 0, 0]),
        ([0], None),
        ([0, 0], None),
        ([0, 0, 0], None),
    ],
)
def test_unit(array, array_unit_expected):

    if array_unit_expected is None:
        with pytest.raises(ValueError, match="The magnitude must not be zero."):
            Vector(array).unit()

    else:
        assert Vector(array).unit().is_close(array_unit_expected)


@pytest.mark.parametrize(
    ("array", "kwargs", "bool_expected"),
    [
        ([0, 0], {}, True),
        ([0, 0, 0], {}, True),
        ([0, 1], {}, False),
        # The tolerance affects the output.
        ([0, 0, 1e-4], {}, False),
        ([0, 0, 1e-4], {'abs_tol': 1e-3}, True),
        ([0, 0, 0, 0], {}, True),
        ([7, 0, 2, 0], {}, False),
    ],
)
def test_is_zero(array, kwargs, bool_expected):

    assert Vector(array).is_zero(**kwargs) == bool_expected


@pytest.mark.parametrize(
    ("array_u", "array_v", "similarity_expected"),
    [
        ([1, 0], [1, 0], 1),
        ([1, 0], [0, 1], 0),
        ([1, 0], [-1, 0], -1),
        ([1, 0], [0, -1], 0),
        ([1, 0], [1, 1], math.sqrt(2) / 2),
        ([1, 0], [-1, 1], -math.sqrt(2) / 2),
        ([1, 0], [-1, -1], -math.sqrt(2) / 2),
        ([1, 0], [1, -1], math.sqrt(2) / 2),
        ([1, 0], [0.5, math.sqrt(3) / 2], 0.5),
        ([1, 0], [math.sqrt(3) / 2, 0.5], math.sqrt(3) / 2),
        ([1, 1], [0, 0], None),
        ([0, 0], [1, 1], None),
    ],
)
def test_cosine_similarity(array_u, array_v, similarity_expected):

    if similarity_expected is None:
        with pytest.raises(ValueError, match="The vectors must have non-zero magnitudes."):
            Vector(array_u).cosine_similarity(array_v)

    else:
        similarity = Vector(array_u).cosine_similarity(array_v)
        assert math.isclose(similarity, similarity_expected)


@pytest.mark.parametrize(
    ("array_u", "array_v", "angle_expected"),
    [
        ([1, 0], [1, 0], 0),
        ([1, 0], [math.sqrt(3) / 2, 0.5], np.pi / 6),
        ([1, 0], [1, 1], np.pi / 4),
        ([1, 0], [0, 1], np.pi / 2),
        ([1, 0], [0, -1], np.pi / 2),
        ([1, 0], [-1, 0], np.pi),
        ([1, 0, 0], [0, 1, 0], np.pi / 2),
    ],
)
def test_angle_between(array_u, array_v, angle_expected):
    """Test finding the angle between vectors u and v."""

    angle = Vector(array_u).angle_between(array_v)
    assert math.isclose(angle, angle_expected)


@pytest.mark.parametrize(
    ("array_u", "array_v", "angle_expected"),
    [
        ([1, 0], [1, 0], 0),
        ([1, 0], [1, 1], np.pi / 4),
        ([1, 0], [0, 1], np.pi / 2),
        ([1, 0], [-1, 1], 3 * np.pi / 4),
        ([1, 0], [-1, 0], np.pi),
        ([1, 0], [-1, -1], -3 * np.pi / 4),
        ([1, 0], [0, -1], -np.pi / 2),
        ([1, 0], [1, -1], -np.pi / 4),
        ([1, 1], [0, 1], np.pi / 4),
        ([1, 1], [1, 0], -np.pi / 4),
        ([0], [0], None),
        ([1, 1, 1], [1, 0, 0], None),
        (np.ones(4), np.ones(4), None),
    ],
)
def test_angle_signed(array_u, array_v, angle_expected):

    if angle_expected is None:
        with pytest.raises(ValueError, match="The vectors must be 2D."):
            Vector(array_u).angle_signed(array_v)

    else:
        angle = Vector(array_u).angle_signed(array_v)
        assert math.isclose(angle, angle_expected)


@pytest.mark.parametrize(
    ("array_u", "array_v", "bool_expected"),
    [
        ([1, 0], [0, 1], True),
        ([0, 1], [-1, 0], True),
        ([-1, 0], [0, -1], True),
        ([1, 1], [-1, -1], False),
        ([1, 1], [1, 1], False),
        # The zero vector is perpendicular to all vectors.
        ([0, 0], [-1, 5], True),
        ([0, 0, 0], [1, 1, 1], True),
    ],
)
def test_is_perpendicular(array_u, array_v, bool_expected):
    """Test checking if vector u is perpendicular to vector v."""
    vector_u = Vector(array_u)

    assert vector_u.is_perpendicular(array_v) == bool_expected


@pytest.mark.parametrize(
    ("array_u", "array_v", "bool_expected"),
    [
        ([0, 1], [0, 1], True),
        ([1, 0], [0, 1], False),
        ([0, 1], [4, 0], False),
        ([0, 1], [0, 5], True),
        ([1, 1], [-1, -1], True),
        ([1, 1], [-5, -5], True),
        ([0, 1], [0, -1], True),
        ([0.1, 5, 4], [3, 2, 0], False),
        ([1, 1, 1, 1], [-2, -2, -2, 4], False),
        ([1, 1, 1, 1], [-2, -2, -2, -2], True),
        ([5, 0, -6, 7], [0, 1, 6, 3], False),
        ([6, 0, 1, 0], [-12, 0, -2, 0], True),
        # The zero vector is parallel to all vectors.
        ([0, 0], [1, 1], True),
        ([5, 2], [0, 0], True),
        ([5, -3, 2, 6], [0, 0, 0, 0], True),
    ],
)
def test_is_parallel(array_u, array_v, bool_expected):
    """Test checking if vector u is parallel to vector v."""
    vector_u = Vector(array_u)

    assert vector_u.is_parallel(array_v) == bool_expected


@pytest.mark.parametrize(
    ("array_a", "array_b", "value_expected"),
    [
        ([0, 0], [0, 0], 0),
        ([0, 0], [0, 1], 0),
        ([0, 0], [1, 1], 0),
        ([0, 1], [0, 1], 0),
        ([0, 1], [0, 9], 0),
        ([0, 1], [0, -20], 0),
        ([0, 1], [1, 1], 1),
        ([0, 1], [38, 29], 1),
        ([0, 1], [1, 0], 1),
        ([0, 1], [1, -100], 1),
        ([0, 1], [-1, 1], -1),
        ([0, 1], [-1, 20], -1),
        ([0, 1], [-1, -20], -1),
        ([0, 1], [-5, 50], -1),
    ],
)
def test_side_vector(array_a, array_b, value_expected):

    assert Vector(array_a).side_vector(array_b) == value_expected


@pytest.mark.parametrize(
    ("array_a", "array_b"),
    [
        ([0], [1]),
        ([0, 0, 0], [1, 1, 1]),
        ([0, 0, 0, 0], [1, 1, 1, 1]),
    ],
)
def test_side_vector_failure(array_a, array_b):

    message_expected = "The vectors must be 2D."

    with pytest.raises(ValueError, match=message_expected):
        Vector(array_a).side_vector(array_b)


@pytest.mark.parametrize(
    ("vector_u", "vector_v", "vector_expected"),
    [
        ([1, 1], [1, 0], [1, 0]),
        ([1, 5], [1, 0], [1, 0]),
        ([5, 5], [1, 0], [5, 0]),
        # Scaling v by a non-zero scalar doesn't change the projection.
        ([0, 1], [0, 1], [0, 1]),
        ([0, 1], [0, -5], [0, 1]),
        ([0, 1], [0, 15], [0, 1]),
        # The projection is the zero vector if u and v are perpendicular.
        ([1, 0], [0, 1], [0, 0]),
        ([5, 0], [0, 9], [0, 0]),
        # The projection of the zero vector onto v is the zero vector.
        ([0, 0], [0, 1], [0, 0]),
    ],
)
def test_project_vector(vector_u, vector_v, vector_expected):
    """Test projecting vector u onto vector v."""

    vector_u_projected = Vector(vector_v).project_vector(vector_u)

    assert vector_u_projected.is_close(vector_expected)


@pytest.mark.parametrize(
    ("array", "array_expected"),
    [
        ([1], [-1]),
        ([5], [-1]),
        ([-5], [1]),
        ([0, 1], [1, 0]),
        ([1, 0], [0, 1]),
        ([2, 0], [0, 1]),
        ([5, 0], [0, 1]),
        ([0, 2], [1, 0]),
        ([0, 5], [1, 0]),
        ([0, 0, 1], [1, 0, 0]),
        ([1, 0, 0], [0, 1, 0]),
        ([1, 0, 1], [1, 0, 0]),
        ([1, 1, 0], [1, 0, 0]),
        ([0, 0, 1, 1], [1, 0, 0, 0]),
        ([5, 0, 1, 1], [1, 0, 0, 0]),
    ],
)
def test_different_direction(array, array_expected):

    vector = Vector(array)
    vector_expected = Vector(array_expected)

    assert vector.different_direction().is_equal(vector_expected)


@pytest.mark.parametrize(
    "array",
    [
        ([0]),
        ([0] * 2),
        ([0] * 3),
        ([0] * 4),
    ],
)
def test_different_direction_failure(array):

    message_expected = "The vector must not be the zero vector."

    with pytest.raises(ValueError, match=message_expected):
        Vector(array).different_direction()



import math

from hypothesis import given

from skspatial.objects import Point
from skspatial.objects import Vector
from tests.property.strategies import arrays_fixed
from tests.property.strategies import consistent_dim


@given(consistent_dim(2 * [arrays_fixed]))
def test_add_subtract(arrays):

    array_point, array_vector = arrays

    point = Point(array_point)
    vector = Vector(array_vector)

    point_2 = point + array_vector
    assert math.isclose(point.distance_point(point_2), vector.norm())

    point_3 = point_2 - array_vector
    assert point.is_close(point_3)



import math
from math import sqrt

import numpy as np
import pytest

from skspatial.objects import Line
from skspatial.objects import Points
from skspatial.objects import Sphere

POINT_MUST_BE_3D = "The point must be 3D."
RADIUS_MUST_BE_POSITIVE = "The radius must be positive."


@pytest.mark.parametrize(
    ("point", "radius", "message_expected"),
    [
        ([0], 1, POINT_MUST_BE_3D),
        ([0, 0], 1, POINT_MUST_BE_3D),
        ([0, 0, 0, 0], 1, POINT_MUST_BE_3D),
        ([1, 2, 3, 4], 1, POINT_MUST_BE_3D),
        ([0, 0, 0], 0, RADIUS_MUST_BE_POSITIVE),
        ([0, 0, 0], -1, RADIUS_MUST_BE_POSITIVE),
        ([0, 0, 0], -5, RADIUS_MUST_BE_POSITIVE),
    ],
)
def test_failure(point, radius, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        Sphere(point, radius)


@pytest.mark.parametrize(
    ("radius", "surface_area_expected", "volume_expected"),
    [
        (1, 4 * np.pi, 4 / 3 * np.pi),
        (2, 16 * np.pi, 32 / 3 * np.pi),
        (3, 36 * np.pi, 36 * np.pi),
        (4.5, 81 * np.pi, 121.5 * np.pi),
        (10, 400 * np.pi, 4000 / 3 * np.pi),
    ],
)
def test_surface_area_volume(radius, surface_area_expected, volume_expected):

    sphere = Sphere([0, 0, 0], radius)

    assert math.isclose(sphere.surface_area(), surface_area_expected)
    assert math.isclose(sphere.volume(), volume_expected)


@pytest.mark.parametrize(
    ("sphere", "point", "dist_expected"),
    [
        (Sphere([0, 0, 0], 1), [0, 0, 0], 1),
        (Sphere([0, 0, 0], 1), [1, 0, 0], 0),
        (Sphere([0, 0, 0], 1), [0, -1, 0], 0),
        (Sphere([0, 0, 0], 2), [0, 0, 0], 2),
        (Sphere([0, 0, 0], 1), [1, 1, 1], math.sqrt(3) - 1),
        (Sphere([0, 0, 0], 2), [1, 1, 1], 2 - math.sqrt(3)),
        (Sphere([1, 0, 0], 2), [0, 0, 0], 1),
    ],
)
def test_distance_point(sphere, point, dist_expected):

    assert math.isclose(sphere.distance_point(point), dist_expected)


@pytest.mark.parametrize(
    ("sphere", "point", "bool_expected"),
    [
        (Sphere([0, 0, 0], 1), [1, 0, 0], True),
        (Sphere([0, 0, 0], 1), [0, 1, 0], True),
        (Sphere([0, 0, 0], 1), [0, 0, 1], True),
        (Sphere([0, 0, 0], 1), [-1, 0, 0], True),
        (Sphere([0, 0, 0], 1), [0, -1, 0], True),
        (Sphere([0, 0, 0], 1), [0, 0, -1], True),
        (Sphere([0, 0, 0], 1), [1, 1, 0], False),
        (Sphere([1, 0, 0], 1), [1, 0, 0], False),
        (Sphere([1, 0, 0], 1), [2, 0, 0], True),
        (Sphere([0, 0, 0], 2), [0, 2, 0], True),
        (Sphere([0, 0, 0], math.sqrt(3)), [1, 1, 1], True),
    ],
)
def test_contains_point(sphere, point, bool_expected):

    assert sphere.contains_point(point) == bool_expected


@pytest.mark.parametrize(
    ("sphere", "point", "point_expected"),
    [
        (Sphere([0, 0, 0], 1), [1, 0, 0], [1, 0, 0]),
        (Sphere([0, 0, 0], 2), [1, 0, 0], [2, 0, 0]),
        (Sphere([0, 0, 0], 0.1), [1, 0, 0], [0.1, 0, 0]),
        (Sphere([-1, 0, 0], 1), [1, 0, 0], [0, 0, 0]),
        (Sphere([0, 0, 0], 1), [1, 1, 1], math.sqrt(3) / 3 * np.ones(3)),
        (Sphere([0, 0, 0], 3), [1, 1, 1], math.sqrt(3) * np.ones(3)),
    ],
)
def test_project_point(sphere, point, point_expected):

    point_projected = sphere.project_point(point)
    assert point_projected.is_close(point_expected)


@pytest.mark.parametrize(
    ("sphere", "point"),
    [
        (Sphere([0, 0, 0], 1), [0, 0, 0]),
        (Sphere([0, 0, 0], 5), [0, 0, 0]),
        (Sphere([5, 2, -6], 5), [5, 2, -6]),
    ],
)
def test_project_point_failure(sphere, point):

    message_expected = "The point must not be the center of the circle or sphere."

    with pytest.raises(ValueError, match=message_expected):
        sphere.project_point(point)


@pytest.mark.parametrize(
    ("points", "sphere_expected"),
    [
        ([[1, 0, 0], [-1, 0, 0], [0, 1, 0], [0, 0, 1]], Sphere(point=[0, 0, 0], radius=1)),
        ([[2, 0, 0], [-2, 0, 0], [0, 2, 0], [0, 0, 2]], Sphere(point=[0, 0, 0], radius=2)),
        ([[1, 0, 1], [0, 1, 1], [1, 2, 1], [1, 1, 2]], Sphere(point=[1, 1, 1], radius=1)),
    ],
)
def test_best_fit(points, sphere_expected):

    points = Points(points)
    sphere_fit = Sphere.best_fit(points)

    assert sphere_fit.point.is_close(sphere_expected.point, abs_tol=1e-9)
    assert math.isclose(sphere_fit.radius, sphere_expected.radius)


@pytest.mark.parametrize(
    ("points", "message_expected"),
    [
        ([[1, 0], [-1, 0], [0, 1], [0, 0]], "The points must be 3D."),
        ([[2, 0, 0], [-2, 0, 0], [0, 2, 0]], "There must be at least 4 points."),
        ([[1, 0, 0], [-1, 0, 0], [0, 1, 0], [0, -1, 0]], "The points must not be in a plane."),
    ],
)
def test_best_fit_failure(points, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        Sphere.best_fit(points)


@pytest.mark.parametrize(
    ("sphere", "line", "point_a_expected", "point_b_expected"),
    [
        (Sphere([0, 0, 0], 1), Line([0, 0, 0], [1, 0, 0]), [-1, 0, 0], [1, 0, 0]),
        (
            Sphere([0, 0, 0], 1),
            Line([0, 0, 0], [1, 1, 0]),
            -sqrt(2) / 2 * np.array([1, 1, 0]),
            sqrt(2) / 2 * np.array([1, 1, 0]),
        ),
        (
            Sphere([0, 0, 0], 1),
            Line([0, 0, 0], [1, 1, 1]),
            -sqrt(3) / 3 * np.ones(3),
            sqrt(3) / 3 * np.ones(3),
        ),
        (Sphere([1, 0, 0], 1), Line([0, 0, 0], [1, 0, 0]), [0, 0, 0], [2, 0, 0]),
        (Sphere([0, 0, 0], 1), Line([1, 0, 0], [0, 0, 1]), [1, 0, 0], [1, 0, 0]),
    ],
)
def test_intersect_line(sphere, line, point_a_expected, point_b_expected):

    point_a, point_b = sphere.intersect_line(line)

    assert point_a.is_close(point_a_expected)
    assert point_b.is_close(point_b_expected)


@pytest.mark.parametrize(
    ("sphere", "line"),
    [
        (Sphere([0, 0, 0], 1), Line([0, 0, 2], [1, 0, 0])),
        (Sphere([0, 0, 0], 1), Line([0, 0, -2], [1, 0, 0])),
        (Sphere([0, 2, 0], 1), Line([0, 0, 0], [1, 0, 0])),
        (Sphere([0, -2, 0], 1), Line([0, 0, 0], [1, 0, 0])),
        (Sphere([5, 0, 0], 1), Line([0, 0, 0], [1, 1, 1])),
    ],
)
def test_intersect_line_failure(sphere, line):

    message_expected = "The line does not intersect the sphere."

    with pytest.raises(ValueError, match=message_expected):
        sphere.intersect_line(line)


@pytest.mark.parametrize(
    ("sphere", "n_angles", "points_expected"),
    [
        (Sphere([0, 0, 0], 1), 1, [[0, 0, 1]]),
        (Sphere([0, 0, 0], 1), 2, [[0, 0, -1], [0, 0, 1]]),
        (Sphere([0, 0, 0], 1), 3, [[0, -1, 0], [0, 0, -1], [0, 0, 1], [0, 1, 0]]),
        (Sphere([0, 0, 0], 2), 3, [[0, -2, 0], [0, 0, -2], [0, 0, 2], [0, 2, 0]]),
        (Sphere([1, 0, 0], 1), 3, [[1, -1, 0], [1, 0, -1], [1, 0, 1], [1, 1, 0]]),
        (Sphere([1, 1, 1], 1), 3, [[1, 0, 1], [1, 1, 0], [1, 1, 2], [1, 2, 1]]),
    ],
)
def test_to_points(sphere, n_angles, points_expected):

    array_rounded = sphere.to_points(n_angles=n_angles).round(3)
    points_unique = Points(array_rounded).unique()

    assert points_unique.is_close(points_expected)



import math

import numpy as np
import pytest
from hypothesis import assume
from hypothesis import given

from skspatial.objects import Point
from skspatial.objects import Vector
from tests.property.constants import ATOL
from tests.property.strategies import arrays
from tests.property.strategies import arrays_fixed
from tests.property.strategies import arrays_fixed_nonzero
from tests.property.strategies import arrays_nonzero
from tests.property.strategies import consistent_dim
from tests.property.strategies import floats


@given(consistent_dim(2 * [arrays_fixed]))
def test_from_points(arrays):

    array_a, array_b = arrays

    point_a = Point(array_a)
    vector_ab = Vector.from_points(array_a, array_b)

    assert (point_a + vector_ab).is_close(array_b)


@given(arrays_nonzero)
def test_unit(array):

    vector = Vector(array)
    vector_unit = vector.unit()

    assert math.isclose(vector_unit.norm(), 1)
    assert (vector.norm() * vector_unit).is_close(array)

    assert vector_unit.is_parallel(vector)

    angle = vector.angle_between(vector_unit)
    assert math.isclose(angle, 0, abs_tol=ATOL)


@given(arrays)
def test_add_subtract(array):
    vector = Vector(array)
    assert (vector + array - array).is_close(array)


@given(arrays_nonzero, floats)
def test_scale(array, scalar):

    assume(abs(scalar) > ATOL)

    vector = Vector(array)
    vector_scaled = scalar * vector

    assert vector_scaled.is_parallel(array)

    angle = vector_scaled.angle_between(array)

    if scalar > 0:
        assert math.isclose(angle, 0, abs_tol=ATOL)
    else:
        assert math.isclose(angle, np.pi, rel_tol=1e-6)


@given(consistent_dim(2 * [arrays_fixed_nonzero]))
def test_two_vectors(arrays):

    array_a, array_b = arrays
    vector_a = Vector(array_a)

    is_perpendicular = vector_a.is_perpendicular(array_b)
    is_parallel = vector_a.is_parallel(array_b)

    # Two non-zero vectors cannot be both perpendicular and parallel.
    assert not (is_perpendicular and is_parallel)

    angle = vector_a.angle_between(array_b)

    if is_perpendicular:
        assert math.isclose(angle, np.pi / 2)

    if is_parallel:
        assert math.isclose(angle, 0, abs_tol=ATOL) or math.isclose(angle, np.pi, rel_tol=1e-6)

    # The zero vector is perpendicular and parallel to any other vector.
    vector_zero = np.zeros(vector_a.size)
    assert vector_a.is_perpendicular(vector_zero)
    assert vector_a.is_parallel(vector_zero)

    # The angle with the zero vector is undefined.
    message_expected = "The vectors must have non-zero magnitudes."

    with pytest.raises(ValueError, match=message_expected):
        vector_a.angle_between(vector_zero)

    # The projection of vector B onto A is parallel to A.
    vector_b_projected = vector_a.project_vector(array_b)
    assert vector_a.is_parallel(vector_b_projected)

    # The projection is zero if vectors A and B are perpendicular.
    if is_perpendicular:
        assert vector_b_projected.is_zero(abs_tol=ATOL)



from hypothesis import given

from tests.property.strategies import triangles


@given(triangles(2))
def test_orthocenter(triangle):

    point_orthocenter = triangle.orthocenter()

    alt_a = triangle.altitude('A')
    alt_b = triangle.altitude('B')
    alt_c = triangle.altitude('C')

    assert all(alt.contains_point(point_orthocenter, abs_tol=0.1) for alt in [alt_a, alt_b, alt_c])



from hypothesis import given
from hypothesis.strategies import integers
from hypothesis.strategies import tuples

from tests.property.constants import ATOL
from tests.property.strategies import planes
from tests.property.strategies import spheres


limits = tuples(integers(-10, 10), integers(-10, 10))


@given(planes(dim=3), limits, limits)
def test_plane_points(plane, lims_x, lims_y):

    points = plane.to_points(lims_x=lims_x, lims_y=lims_y)

    # Test that all the points are on the plane.
    assert all(plane.contains_point(point, abs_tol=ATOL) for point in points)


@given(spheres(), integers(1, 30))
def test_sphere_points(sphere, n_angles):

    points = sphere.to_points(n_angles=n_angles)

    # Test that all the points are on the plane.
    assert all(sphere.contains_point(point, abs_tol=ATOL) for point in points)



from dataclasses import dataclass
from math import atan
from math import degrees
from math import isclose
from math import radians
from math import sqrt

import pytest

from skspatial.objects import Line
from skspatial.objects import Triangle
from skspatial.typing import array_like


@dataclass
class TriangleTester:
    """Triangle object for unit testing."""

    points: tuple

    area: float
    perimeter: float

    lengths: tuple
    angles: tuple
    altitudes: tuple

    normal: array_like
    centroid: array_like
    orthocenter: array_like

    classification: str
    is_right: bool


list_test_cases = [
    TriangleTester(
        points=[[0, 0], [1, 0], [0, 1]],
        area=0.5,
        perimeter=2 + sqrt(2),
        lengths=(sqrt(2), 1, 1),
        angles=(90, 45, 45),
        centroid=[1 / 3, 1 / 3],
        orthocenter=[0, 0],
        normal=[0, 0, 1],
        classification='isosceles',
        is_right=True,
        altitudes=(Line([0, 0], [0.5, 0.5]), Line([1, 0], [-1, 0]), Line([0, 1], [0, -1])),
    ),
    TriangleTester(
        points=[[0, 0], [1, 1], [2, 0]],
        area=1,
        perimeter=2 + 2 * sqrt(2),
        lengths=(sqrt(2), 2, sqrt(2)),
        angles=(45, 90, 45),
        centroid=[1, 1 / 3],
        orthocenter=[1, 1],
        normal=[0, 0, -2],
        classification='isosceles',
        is_right=True,
        altitudes=(Line([0, 0], [1, 1]), Line([1, 1], [0, -1]), Line([2, 0], [-1, 1])),
    ),
    TriangleTester(
        points=[[0, 0], [1, 0], [0.5, sqrt(3) / 2]],
        area=sqrt(3) / 4,
        perimeter=3,
        lengths=(1, 1, 1),
        angles=(60, 60, 60),
        centroid=[0.5, sqrt(3) / 6],
        orthocenter=[0.5, sqrt(3) / 6],
        normal=[0, 0, sqrt(3) / 2],
        classification='equilateral',
        is_right=False,
        altitudes=(
            Line([0, 0], [0.75, sqrt(3) / 4]),
            Line([1, 0], [-0.75, sqrt(3) / 4]),
            Line([0.5, sqrt(3) / 2], [0, -sqrt(3) / 2]),
        ),
    ),
    TriangleTester(
        points=[[0, 0], [1, 0], [0, 2]],
        area=1,
        perimeter=3 + sqrt(5),
        lengths=(sqrt(5), 2, 1),
        angles=(90, degrees(atan(2)), degrees(atan(1 / 2))),
        centroid=[1 / 3, 2 / 3],
        orthocenter=[0, 0],
        normal=[0, 0, 2],
        classification='scalene',
        is_right=True,
        altitudes=(Line([0, 0], [0.8, 0.4]), Line([1, 0], [-1, 0]), Line([0, 2], [0, -2])),
    ),
    TriangleTester(
        points=[[0, 0], [3, 0], [0, 4]],
        area=6,
        perimeter=12,
        lengths=(5, 4, 3),
        angles=(90, degrees(atan(4 / 3)), degrees(atan(3 / 4))),
        centroid=[1, 4 / 3],
        orthocenter=[0, 0],
        normal=[0, 0, 12],
        classification='scalene',
        is_right=True,
        altitudes=(Line([0, 0], [1.92, 1.44]), Line([3, 0], [-3, 0]), Line([0, 4], [0, -4])),
    ),
]


@pytest.mark.parametrize('test_case', list_test_cases)
def test_triangle(test_case):

    triangle = Triangle(*test_case.points)

    assert triangle.area() == test_case.area
    assert triangle.perimeter() == test_case.perimeter

    points_a = triangle.multiple('point', 'ABC')
    points_b = test_case.points
    assert all(a.is_equal(b) for a, b in zip(points_a, points_b))

    lengths_a = triangle.multiple('length', 'abc')
    lengths_b = test_case.lengths
    assert all(isclose(a, b) for a, b in zip(lengths_a, lengths_b))

    angles_a = triangle.multiple('angle', 'ABC')
    angles_b = tuple(map(radians, test_case.angles))
    assert all(isclose(a, b, abs_tol=1e-3) for a, b in zip(angles_a, angles_b))

    assert triangle.normal().is_close(test_case.normal)
    assert triangle.centroid().is_close(test_case.centroid)
    assert triangle.orthocenter().is_close(test_case.orthocenter)

    altitudes_a = triangle.multiple('altitude', 'ABC')
    altitudes_b = test_case.altitudes
    print(altitudes_a, altitudes_b)
    assert all(a.is_close(b, abs_tol=1e-3) for a, b in zip(altitudes_a, altitudes_b))

    assert triangle.classify() == test_case.classification
    assert triangle.is_right() == test_case.is_right


@pytest.mark.parametrize(
    ("array_a", "array_b", "array_c"),
    [([1], [1, 0], [1, 0]), ([1, 0, 0], [1, 0], [1, 0]), ([1, 0], [1, 0], [1, 0, 0]), ([1, 0, 0], [1, 0], [1, 0, 0])],
)
def test_failure_different_dimensions(array_a, array_b, array_c):

    with pytest.raises(ValueError, match="The points must have the same dimension."):
        Triangle(array_a, array_b, array_c)


@pytest.mark.parametrize(
    ("array_a", "array_b", "array_c"),
    [
        ([1], [2], [3]),
        ([1, 0], [1, 0], [1, 0]),
        ([1, 2, 3], [1, 2, 3], [1, 2, 3]),
        ([1, 2, 3], [1, 2, 3], [2, 3, -10]),  # Two points are the same.
        ([1, 2, 3], [4, 5, 6], [7, 8, 9]),
        ([1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]),
    ],
)
def test_failure_collinear_points(array_a, array_b, array_c):

    with pytest.raises(ValueError, match="The points must not be collinear."):
        Triangle(array_a, array_b, array_c)


@pytest.fixture()
def basic_triangle():

    return Triangle([0, 0], [0, 1], [1, 0])


@pytest.mark.parametrize("string", ['a', 'b', 'c', 'd', 'D'])
def test_failure_point(basic_triangle, string):

    message = "The vertex must be 'A', 'B', or 'C'."

    with pytest.raises(ValueError, match=message):
        basic_triangle.point(string)

    with pytest.raises(ValueError, match=message):
        basic_triangle.angle(string)

    with pytest.raises(ValueError, match=message):
        basic_triangle.altitude(string)


@pytest.mark.parametrize("string", ['A', 'B', 'C', 'D'])
def test_failure_line(basic_triangle, string):

    message = "The side must be 'a', 'b', or 'c'."

    with pytest.raises(ValueError, match=message):
        basic_triangle.line(string)

    with pytest.raises(ValueError, match=message):
        basic_triangle.length(string)



import math

import numpy as np
import pytest

from skspatial.objects import Circle
from skspatial.objects import Line
from skspatial.objects import Points


@pytest.mark.parametrize(
    ("point", "radius", "message_expected"),
    [
        ([0, 0, 0], 1, "The point must be 2D"),
        ([1, 2, 3], 1, "The point must be 2D"),
        ([0, 0], 0, "The radius must be positive"),
        ([0, 0], -1, "The radius must be positive"),
        ([0, 0], -5, "The radius must be positive"),
    ],
)
def test_failure(point, radius, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        Circle(point, radius)


@pytest.mark.parametrize(
    ("radius", "circumference_expected", "area_expected"),
    [
        (1, 2 * np.pi, np.pi),
        (2, 4 * np.pi, 4 * np.pi),
        (3, 6 * np.pi, 9 * np.pi),
        (4.5, 9 * np.pi, 20.25 * np.pi),
        (10, 20 * np.pi, 100 * np.pi),
    ],
)
def test_circumference_area(radius, circumference_expected, area_expected):

    circle = Circle([0, 0], radius)

    assert math.isclose(circle.circumference(), circumference_expected)
    assert math.isclose(circle.area(), area_expected)


@pytest.mark.parametrize(
    ("circle", "point", "dist_expected"),
    [
        (Circle([0, 0], 1), [0, 0], 1),
        (Circle([0, 0], 1), [0.5, 0], 0.5),
        (Circle([0, 0], 1), [1, 0], 0),
        (Circle([0, 0], 1), [0, 1], 0),
        (Circle([0, 0], 1), [-1, 0], 0),
        (Circle([0, 0], 1), [0, -1], 0),
        (Circle([0, 0], 1), [2, 0], 1),
        (Circle([0, 0], 1), [1, 1], math.sqrt(2) - 1),
        (Circle([1, 1], 1), [0, 0], math.sqrt(2) - 1),
        (Circle([0, 0], 2), [0, 5], 3),
    ],
)
def test_distance_point(circle, point, dist_expected):

    assert math.isclose(circle.distance_point(point), dist_expected)


@pytest.mark.parametrize(
    ("circle", "point", "bool_expected"),
    [
        (Circle([0, 0], 1), [1, 0], True),
        (Circle([0, 0], 1), [0, 1], True),
        (Circle([0, 0], 1), [-1, 0], True),
        (Circle([0, 0], 1), [0, -1], True),
        (Circle([0, 0], 1), [0, 0], False),
        (Circle([0, 0], 1), [1, 1], False),
        (Circle([0, 0], 2), [1, 0], False),
        (Circle([1, 0], 1), [1, 0], False),
        (Circle([0, 0], math.sqrt(2)), [1, 1], True),
    ],
)
def test_contains_point(circle, point, bool_expected):

    assert circle.contains_point(point) == bool_expected


@pytest.mark.parametrize(
    ("circle", "point", "point_expected"),
    [
        (Circle([0, 0], 1), [1, 0], [1, 0]),
        (Circle([0, 0], 1), [2, 0], [1, 0]),
        (Circle([0, 0], 1), [-2, 0], [-1, 0]),
        (Circle([0, 0], 1), [0, 2], [0, 1]),
        (Circle([0, 0], 1), [0, -2], [0, -1]),
        (Circle([0, 0], 5), [0, -2], [0, -5]),
        (Circle([0, 1], 5), [0, -2], [0, -4]),
        (Circle([0, 0], 1), [1, 1], math.sqrt(2) / 2 * np.ones(2)),
        (Circle([0, 0], 2), [1, 1], math.sqrt(2) * np.ones(2)),
    ],
)
def test_project_point(circle, point, point_expected):

    point_projected = circle.project_point(point)
    assert point_projected.is_close(point_expected)


@pytest.mark.parametrize(
    ("circle", "point"),
    [
        (Circle([0, 0], 1), [0, 0]),
        (Circle([0, 0], 5), [0, 0]),
        (Circle([7, -1], 5), [7, -1]),
    ],
)
def test_project_point_failure(circle, point):

    message_expected = "The point must not be the center of the circle or sphere."

    with pytest.raises(ValueError, match=message_expected):
        circle.project_point(point)


@pytest.mark.parametrize(
    ("points", "circle_expected"),
    [
        ([[1, 1], [2, 2], [3, 1]], Circle(point=[2, 1], radius=1)),
        ([[2, 0], [-2, 0], [0, 2]], Circle(point=[0, 0], radius=2)),
        ([[1, 0], [0, 1], [1, 2]], Circle(point=[1, 1], radius=1)),
    ],
)
def test_best_fit(points, circle_expected):

    points = Points(points)
    circle_fit = Circle.best_fit(points)

    assert circle_fit.point.is_close(circle_expected.point, abs_tol=1e-9)
    assert math.isclose(circle_fit.radius, circle_expected.radius)


@pytest.mark.parametrize(
    ("points", "message_expected"),
    [
        ([[1, 0, 0], [-1, 0, 0], [0, 1, 0]], "The points must be 2D."),
        ([[2, 0], [-2, 0]], "There must be at least 3 points."),
        ([[0, 0], [1, 1], [2, 2]], "The points must not be collinear."),
    ],
)
def test_best_fit_failure(points, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        Circle.best_fit(points)


@pytest.mark.parametrize(
    ("circle", "line", "point_a_expected", "point_b_expected"),
    [
        (Circle([0, 0], 1), Line([0, 0], [1, 0]), [-1, 0], [1, 0]),
        (Circle([0, 0], 1), Line([0, 0], [0, 1]), [0, -1], [0, 1]),
        (Circle([0, 0], 1), Line([0, 1], [1, 0]), [0, 1], [0, 1]),
        (
            Circle([0, 0], 1),
            Line([0, 0.5], [1, 0]),
            [-math.sqrt(3) / 2, 0.5],
            [math.sqrt(3) / 2, 0.5],
        ),
        (Circle([1, 0], 1), Line([0, 0], [1, 0]), [0, 0], [2, 0]),
        (Circle([1.5, 0], 1), Line([0, 0], [1, 0]), [0.5, 0], [2.5, 0]),
    ],
)
def test_intersect_line(circle, line, point_a_expected, point_b_expected):

    point_a, point_b = circle.intersect_line(line)

    assert point_a.is_close(point_a_expected)
    assert point_b.is_close(point_b_expected)


@pytest.mark.parametrize(
    ("circle", "line"),
    [
        (Circle([0, 0], 1), Line([0, 2], [1, 0])),
        (Circle([0, 0], 1), Line([0, -2], [1, 0])),
        (Circle([0, 0], 1), Line([2, 0], [0, 1])),
        (Circle([0, 0], 1), Line([3, 0], [1, 1])),
        (Circle([0, 0], 0.5), Line([0, 1], [1, 1])),
        (Circle([0, 1], 0.5), Line([0, 0], [1, 0])),
        (Circle([5, 2], 1), Line([2, -1], [1, 0])),
    ],
)
def test_intersect_line_failure(circle, line):

    message_expected = "The line does not intersect the circle."

    with pytest.raises(ValueError, match=message_expected):
        circle.intersect_line(line)



from math import isclose
from math import sqrt

import pytest

from skspatial._functions import _solve_quadratic

A_MUST_BE_NON_ZERO = "The coefficient `a` must be non-zero."
DISCRIMINANT_MUST_NOT_BE_NEGATIVE = "The discriminant must not be negative."


@pytest.mark.parametrize(
    ("a", "b", "c", "x1_expected", "x2_expected"),
    [
        (1, 0, 0, 0, 0),
        (-1, 0, 0, 0, 0),
        (-1, 1, 0, 1, 0),
        (1, -1, -1, (1 - sqrt(5)) / 2, (1 + sqrt(5)) / 2),
    ],
)
def test_solve_quadratic(a, b, c, x1_expected, x2_expected):

    x1, x2 = _solve_quadratic(a, b, c)

    assert isclose(x1, x1_expected)
    assert isclose(x2, x2_expected)


@pytest.mark.parametrize(
    ("a", "b", "c", "message_expected"),
    [
        (0, 0, 0, A_MUST_BE_NON_ZERO),
        (0, 1, 1, A_MUST_BE_NON_ZERO),
        (1, 0, 1, DISCRIMINANT_MUST_NOT_BE_NEGATIVE),
        (1, 2, 2, DISCRIMINANT_MUST_NOT_BE_NEGATIVE),
        (1, -2, 2, DISCRIMINANT_MUST_NOT_BE_NEGATIVE),
        (-1, 0, -1, DISCRIMINANT_MUST_NOT_BE_NEGATIVE),
    ],
)
def test_solve_quadratic_failure(a, b, c, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        _solve_quadratic(a, b, c)



import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from numpy.testing import assert_array_equal

from skspatial.objects import Point
from skspatial.objects import Points


@pytest.mark.parametrize(
    "array",
    [
        [0],
        [5],
        [0, 1],
        [0, 1, 2],
    ],
)
def test_failure(array):

    message_expected = "The array must be 2D."

    with pytest.raises(ValueError, match=message_expected):
        Points(array)


@pytest.mark.parametrize(
    ("points", "dim_expected"),
    [
        (Points([[0, 0], [1, 1]]), 2),
        (Points([[0, 0], [0, 0], [0, 0]]), 2),
        (Points([[0, 0, 1], [1, 2, 1]]), 3),
        (Points([[4, 3, 9, 1], [3, 7, 8, 1]]), 4),
    ],
)
def test_dimension(points, dim_expected):

    assert points.dimension == dim_expected


@pytest.mark.parametrize(
    ("points", "dim", "points_expected"),
    [
        (Points([[0, 0], [1, 1]]), 3, Points([[0, 0, 0], [1, 1, 0]])),
        (Points([[0, 0], [1, 1]]), 4, Points([[0, 0, 0, 0], [1, 1, 0, 0]])),
        # The same dimension is allowed (nothing is changed).
        (Points([[0, 0, 0], [1, 1, 1]]), 3, Points([[0, 0, 0], [1, 1, 1]])),
    ],
)
def test_set_dimension(points, dim, points_expected):

    assert_array_equal(points.set_dimension(dim), points_expected)


def test_dimension_of_slice():

    points = Points([[0, 0, 0], [0, 0, 0]])

    assert points.dimension == 3
    assert points[:, :3].dimension == 3
    assert points[:, :2].dimension == 2
    assert points[:, :1].dimension == 1

    assert points[:0, :].dimension == 3
    assert points[:1, :].dimension == 3
    assert points[:2, :].dimension == 3

    # The dimension value is None here because the points are no longer a 2D array,
    # so the normal method of finding the dimension fails.
    assert points[:, 0].dimension is None
    assert points[0, :].dimension is None


@pytest.mark.parametrize(
    ("array_points", "array_centered_expected", "centroid_expected"),
    [
        ([[0, 1]], [[0, 0]], [0, 1]),
        ([[1, 1], [2, 2]], [[-0.5, -0.5], [0.5, 0.5]], [1.5, 1.5]),
        ([[0, 0], [2, 2]], [[-1, -1], [1, 1]], [1, 1]),
        (
            [[1, 2], [-1.3, 11], [24, 5], [7, 3]],
            [[-6.675, -3.25], [-8.975, 5.75], [16.325, -0.25], [-0.675, -2.25]],
            [7.675, 5.25],
        ),
        ([[-2, 0, 2, 5], [4, 1, -3, 2.1]], [[-3, -0.5, 2.5, 1.45], [3, 0.5, -2.5, -1.45]], [1, 0.5, -0.5, 3.55]),
    ],
)
def test_mean_center(array_points, array_centered_expected, centroid_expected):

    points = Points(array_points)
    points_centered, centroid = points.mean_center(return_centroid=True)

    assert isinstance(points_centered, Points)
    assert isinstance(centroid, Point)

    assert_array_almost_equal(points_centered, array_centered_expected)
    assert_array_almost_equal(centroid, centroid_expected)


@pytest.mark.parametrize(
    ("array_points", "array_points_expected"),
    [
        ([[0, 0], [1, 0]], [[0, 0], [1, 0]]),
        ([[0, 0], [1, 1]], [[0, 0], np.sqrt(2) / 2 * np.ones(2)]),
        ([[0, 0], [5, 0], [-1, 0]], [[0, 0], [1, 0], [-0.2, 0]]),
        (9 * np.ones((3, 3)), np.sqrt(3) / 3 * np.ones((3, 3))),
    ],
)
def test_normalize_distance(array_points, array_points_expected):

    points_normalized = Points(array_points).normalize_distance()

    assert_array_almost_equal(points_normalized, array_points_expected)


@pytest.mark.parametrize(
    ("points", "bool_expected"),
    [
        ([[0, 0], [0, 0], [0, 0]], True),
        ([[1, 0], [1, 0], [1, 0]], True),
        ([[0, 0], [0, 1], [0, 2]], True),
        ([[0, 0], [0, 1], [1, 2]], False),
        ([[0, 1], [0, 0], [0, 2]], True),
        ([[0, 0], [-1, 0], [10, 0]], True),
        ([[0, 0], [1, 1], [2, 2], [-4, -4], [5, 5]], True),
        ([[0, 0, 0], [1, 1, 1], [2, 2, 2]], True),
        ([[0, 0, 0], [1, 1, 1], [2, 2, 2.5]], False),
        ([[0, 0, 0], [1, 1, 0], [2, 2, 0], [-4, -4, 10], [5, 5, 0]], False),
    ],
)
def test_are_collinear(points, bool_expected):
    """Test checking if multiple points are collinear."""

    assert Points(points).are_collinear() == bool_expected



import hypothesis.strategies as st
import numpy as np
from hypothesis import assume
from hypothesis import given
from numpy.testing import assert_array_almost_equal

from tests.property.constants import ATOL
from tests.property.strategies import lines
from tests.property.strategies import multi_points


@given(st.data())
def test_mean_center(data):

    dim = data.draw(st.integers(min_value=2, max_value=5))
    points = data.draw(multi_points(dim))

    points_centered = points.mean_center()

    # The centroid of the centered points should be the origin.
    centroid_new = points_centered.centroid()
    origin = np.zeros(points.dimension)

    assert_array_almost_equal(centroid_new, origin)


@given(st.data())
def test_normalize_distance(data):

    dim = data.draw(st.integers(min_value=2, max_value=5))
    points = data.draw(multi_points(dim))

    assume(not np.all(points == 0))

    points_transformed = points.normalize_distance()
    distances_to_points = np.linalg.norm(points_transformed, axis=1)

    assert np.logical_and(distances_to_points >= 0, distances_to_points <= 1 + ATOL).all()


@given(st.data())
def test_transform_points_line(data):

    dim = data.draw(st.integers(min_value=2, max_value=5))
    points = data.draw(multi_points(dim))

    line = data.draw(lines(dim))

    # Transform the points into 1D coordinates.
    coordinates = line.transform_points(points)

    # Project the points onto the line.
    points_projected = np.apply_along_axis(line.project_point, 1, points)

    # Find the signed distances from the line point to the projected points.
    vectors_to_projections = points_projected - line.point
    direction_unit = line.direction.unit()
    distances_signed = np.apply_along_axis(direction_unit.dot, 1, vectors_to_projections)

    # The coordinates from the transformation should be equal to the signed distances.
    assert_array_almost_equal(distances_signed, coordinates)



"""Module for base class of Circle and Sphere."""
import numpy as np

from skspatial._functions import _contains_point
from skspatial.objects._base_spatial import _BaseSpatial
from skspatial.objects.point import Point
from skspatial.objects.vector import Vector
from skspatial.typing import array_like


class _BaseSphere(_BaseSpatial):
    """Private parent class for Circle and Sphere."""

    def __init__(self, point: array_like, radius: float):

        if radius <= 0:
            raise ValueError("The radius must be positive.")

        self.point = Point(point)
        self.radius = radius

        self.dimension = self.point.dimension

    def __repr__(self) -> str:

        name_class = type(self).__name__

        repr_point = np.array_repr(self.point)

        return f"{name_class}(point={repr_point}, radius={self.radius})"

    def distance_point(self, point: array_like) -> np.float64:
        """Return the distance from a point to the circle/sphere."""
        distance_to_center = self.point.distance_point(point)

        return abs(distance_to_center - self.radius)

    def contains_point(self, point: array_like, **kwargs: float) -> bool:
        """Check if the circle/sphere contains a point."""
        return _contains_point(self, point, **kwargs)

    def project_point(self, point: array_like) -> Point:
        """
        Project a point onto the circle or sphere.

        Parameters
        ----------
        point : array_like
            Input point.

        Returns
        -------
        Point
            Point projected onto the circle or sphere.

        Raises
        ------
        ValueError
            If the input point is the center of the circle or sphere.

        Examples
        --------
        >>> from skspatial.objects import Circle

        >>> circle = Circle([0, 0], 1)

        >>> circle.project_point([1, 1]).round(3)
        Point([0.707, 0.707])

        >>> circle.project_point([-6, 3]).round(3)
        Point([-0.894,  0.447])

        >>> circle.project_point([0, 0])
        Traceback (most recent call last):
        ...
        ValueError: The point must not be the center of the circle or sphere.

        >>> from skspatial.objects import Sphere

        >>> Sphere([0, 0, 0], 2).project_point([1, 2, 3]).round(3)
        Point([0.535, 1.069, 1.604])

        """
        if self.point.is_equal(point):
            raise ValueError("The point must not be the center of the circle or sphere.")

        vector_to_point = Vector.from_points(self.point, point)

        return self.point + self.radius * vector_to_point.unit()



import math

from hypothesis import given

from skspatial.objects import Line
from tests.property.constants import ATOL
from tests.property.strategies import consistent_dim
from tests.property.strategies import lines
from tests.property.strategies import points
from tests.property.strategies import vectors_nonzero


@given(consistent_dim([points, vectors_nonzero]))
def test_from_points(objs):

    point_a, vector = objs
    point_b = point_a + vector

    line = Line(point_a, vector)
    line_from_points = Line.from_points(point_a, point_b)

    assert line.is_close(line_from_points, abs_tol=ATOL)

    # The line of best fit should be the same
    # as the line from two points.
    line_fit = Line.best_fit([point_a, point_b])
    assert line_fit.is_close(line_from_points, abs_tol=ATOL)


@given(consistent_dim(2 * [lines]))
def test_two_lines(lines):

    line_a, line_b = lines

    if line_a.direction.is_parallel(line_b.direction, rel_tol=0, abs_tol=0):
        # The lines are parallel, so they must be coplanar.
        assert line_a.is_coplanar(line_b)

    if line_a.is_coplanar(line_b, tol=0) and not line_a.direction.is_parallel(line_b.direction):
        # The lines are coplanar but not parallel, so they must intersect.
        distance = line_a.distance_line(line_b)
        assert math.isclose(distance, 0, abs_tol=ATOL)



"""Module for the Cylinder class."""
from __future__ import annotations

from typing import Optional
from typing import Tuple

import numpy as np
from mpl_toolkits.mplot3d import Axes3D

from skspatial._functions import _solve_quadratic
from skspatial.objects._base_spatial import _BaseSpatial
from skspatial.objects._mixins import _ToPointsMixin
from skspatial.objects.line import Line
from skspatial.objects.plane import Plane
from skspatial.objects.point import Point
from skspatial.objects.vector import Vector
from skspatial.typing import array_like


class Cylinder(_BaseSpatial, _ToPointsMixin):
    """
    A cylinder in space.

    The cylinder is defined by a point at its base, a vector along its axis, and a radius.

    Parameters
    ----------
    point : array_like
        Centre of the cylinder base.
    vector : array_like
        Normal vector of the cylinder base (the vector along the cylinder axis).
        The length of the cylinder is the length of this vector.
    radius : {int, float}
        Radius of the cylinder.
        This is the radius of the circular base.

    Attributes
    ----------
    point : Point
        Centre of the cylinder base.
    vector : Vector
        Normal vector of the cylinder base.
    radius : {int, float}
        Radius of the cylinder.
    dimension : int
        Dimension of the cylinder.

    Raises
    ------
    ValueError
        If the point or vector are not 3D.
        If the vector is all zeros.
        If the radius is zero.

    Examples
    --------
    >>> from skspatial.objects import Cylinder

    >>> Cylinder([0, 0], [1, 0, 0], 1)
    Traceback (most recent call last):
    ...
    ValueError: The point must be 3D.

    >>> Cylinder([0, 0, 0], [1, 0], 1)
    Traceback (most recent call last):
    ...
    ValueError: The vector must be 3D.

    >>> Cylinder([0, 0, 0], [0, 0, 0], 1)
    Traceback (most recent call last):
    ...
    ValueError: The vector must not be the zero vector.

    >>> Cylinder([0, 0, 0], [0, 0, 1], 0)
    Traceback (most recent call last):
    ...
    ValueError: The radius must be positive.

    >>> cylinder = Cylinder([0, 0, 0], [0, 0, 1], 1)

    >>> cylinder
    Cylinder(point=Point([0, 0, 0]), vector=Vector([0, 0, 1]), radius=1)

    >>> cylinder.point
    Point([0, 0, 0])
    >>> cylinder.vector
    Vector([0, 0, 1])
    >>> cylinder.radius
    1
    >>> cylinder.dimension
    3

    """

    def __init__(self, point: array_like, vector: array_like, radius: float):

        self.point = Point(point)
        self.vector = Vector(vector)

        if self.point.dimension != 3:
            raise ValueError("The point must be 3D.")

        if self.vector.dimension != 3:
            raise ValueError("The vector must be 3D.")

        if self.vector.is_zero():
            raise ValueError("The vector must not be the zero vector.")

        if not radius > 0:
            raise ValueError("The radius must be positive.")

        self.radius = radius

        self.dimension = self.point.dimension

    def __repr__(self) -> str:

        repr_point = np.array_repr(self.point)
        repr_vector = np.array_repr(self.vector)

        return f"Cylinder(point={repr_point}, vector={repr_vector}, radius={self.radius})"

    @classmethod
    def from_points(cls, point_a: array_like, point_b: array_like, radius: float) -> Cylinder:
        """
        Instantiate a cylinder from two points and a radius.

        Parameters
        ----------
        point_a, point_b : array_like
            The centres of the two circular ends.
        radius : float
            The cylinder radius.

        Returns
        -------
        Cylinder
            The cylinder defined by the two points and the radius.

        Examples
        --------
        >>> from skspatial.objects import Cylinder

        >>> Cylinder.from_points([0, 0, 0], [0, 0, 1], 1)
        Cylinder(point=Point([0, 0, 0]), vector=Vector([0, 0, 1]), radius=1)

        >>> Cylinder.from_points([0, 0, 0], [0, 0, 2], 1)
        Cylinder(point=Point([0, 0, 0]), vector=Vector([0, 0, 2]), radius=1)

        """
        vector_ab = Vector.from_points(point_a, point_b)

        return cls(point_a, vector_ab, radius)

    def length(self) -> np.float64:
        """
        Return the length of the cylinder.

        This is the length of the vector used to initialize the cylinder.

        Returns
        -------
        np.float64
            Length of the cylinder.

        Examples
        --------
        >>> from skspatial.objects import Cylinder

        >>> Cylinder([0, 0, 0], [0, 0, 1], 1).length()
        1.0

        >>> Cylinder([0, 0, 0], [0, 0, 2], 1).length()
        2.0

        >>> Cylinder([0, 0, 0], [1, 1, 1], 1).length().round(3)
        1.732

        """
        return self.vector.norm()

    def lateral_surface_area(self) -> np.float64:
        """
        Return the lateral surface area of the cylinder.

        Returns
        -------
        np.float64
            Lateral surface area of the cylinder.

        Examples
        --------
        >>> from skspatial.objects import Cylinder

        >>> Cylinder([0, 0, 0], [0, 0, 1], 1).lateral_surface_area().round(3)
        6.283

        >>> Cylinder([0, 0, 0], [0, 0, 1], 2).lateral_surface_area().round(3)
        12.566

        >>> Cylinder([0, 0, 0], [0, 0, 2], 2).lateral_surface_area().round(3)
        25.133

        """
        return 2 * np.pi * self.radius * self.length()

    def surface_area(self) -> np.float64:
        """
        Return the total surface area of the cylinder.

        This is the lateral surface area plus the area of the two circular caps.

        Returns
        -------
        np.float64
            Total surface area of the cylinder.

        Examples
        --------
        >>> from skspatial.objects import Cylinder

        >>> Cylinder([0, 0, 0], [0, 0, 1], 1).surface_area().round(3)
        12.566

        >>> Cylinder([0, 0, 0], [0, 0, 1], 2).surface_area().round(3)
        37.699

        >>> Cylinder([0, 0, 0], [0, 0, 2], 2).surface_area().round(3)
        50.265

        """
        return self.lateral_surface_area() + 2 * np.pi * self.radius**2

    def volume(self) -> np.float64:
        r"""
        Return the volume of the cylinder.

        The volume :math:`V` of a cylinder with radius :math:`r` and length :math:`l` is

        .. math:: V = \pi r^2 l

        Returns
        -------
        np.float64
            Volume of the cylinder.

        Examples
        --------
        >>> from skspatial.objects import Cylinder

        >>> Cylinder([0, 0, 0], [0, 0, 1], 1).volume().round(5)
        3.14159

        The length of the vector sets the length of the cylinder.

        >>> Cylinder([0, 0, 0], [0, 0, 2], 1).volume().round(5)
        6.28319

        """
        return np.pi * self.radius**2 * self.length()

    def is_point_within(self, point: array_like) -> bool:
        """
        Check if a point is within the cylinder.

        This also includes a point on the surface.

        Parameters
        ----------
        point : array_like
            Input point

        Returns
        -------
        bool
            True if the point is within the cylinder.

        Examples
        --------
        >>> from skspatial.objects import Cylinder

        >>> cylinder = Cylinder([0, 0, 0], [0, 0, 1], 1)

        >>> cylinder.is_point_within([0, 0, 0])
        True
        >>> cylinder.is_point_within([0, 0, 1])
        True
        >>> cylinder.is_point_within([0, 0, 2])
        False
        >>> cylinder.is_point_within([0, 0, -1])
        False
        >>> cylinder.is_point_within([1, 0, 0])
        True
        >>> cylinder.is_point_within([0, 1, 0])
        True
        >>> cylinder.is_point_within([1, 1, 0])
        False

        """
        line_axis = Line(self.point, self.vector)
        distance_to_axis = line_axis.distance_point(point)

        within_radius = distance_to_axis <= self.radius

        between_cap_planes = _between_cap_planes(self, point)

        return within_radius and between_cap_planes

    def intersect_line(
        self,
        line: Line,
        n_digits: Optional[int] = None,
        infinite: bool = True,
    ) -> Tuple[Point, Point]:
        """
        Intersect the cylinder with a 3D line.

        By default, this method treats the cylinder as infinite along its axis (i.e., without caps).

        Parameters
        ----------
        line : Line
            Input 3D line.
        n_digits : int, optional
            Additional keywords passed to :func:`round`.
            This is used to round the coefficients of the quadratic equation.
        infinite : bool
            If True, the cylinder is treated as infinite along its axis (i.e., without caps).

        Returns
        -------
        point_a, point_b: Point
            The two intersection points of the line with the cylinder, if they exist.

        Raises
        ------
        ValueError
            If the line is not 3D.
            If the line does not intersect the cylinder at one or two points.

        References
        ----------
        https://mrl.cs.nyu.edu/~dzorin/rendering/lectures/lecture3/lecture3.pdf

        Examples
        --------
        >>> from skspatial.objects import Line, Cylinder

        >>> cylinder = Cylinder([0, 0, 0], [0, 0, 1], 1)
        >>> line = Line([0, 0, 0], [1, 0, 0])


        Intersection with an infinite cylinder.

        >>> cylinder.intersect_line(line)
        (Point([-1.,  0.,  0.]), Point([1., 0., 0.]))

        >>> line = Line([1, 2, 3], [1, 2, 3])

        >>> point_a, point_b = cylinder.intersect_line(line)

        >>> point_a.round(3)
        Point([-0.447, -0.894, -1.342])
        >>> point_b.round(3)
        Point([0.447, 0.894, 1.342])

        >>> cylinder.intersect_line(Line([0, 0], [1, 2]))
        Traceback (most recent call last):
        ...
        ValueError: The line must be 3D.

        >>> cylinder.intersect_line(Line([0, 0, 2], [0, 0, 1]))
        Traceback (most recent call last):
        ...
        ValueError: The line does not intersect the cylinder.

        >>> cylinder.intersect_line(Line([2, 0, 0], [0, 1, 1]))
        Traceback (most recent call last):
        ...
        ValueError: The line does not intersect the cylinder.


        Intersection with a finite cylinder.

        >>> point_a, point_b = cylinder.intersect_line(Line([0, 0, 0], [0, 0, 1]), infinite=False)

        >>> point_a
        Point([0., 0., 0.])
        >>> point_b
        Point([0., 0., 1.])

        >>> cylinder = Cylinder([0, 0, 0], [0, 0, 5], 1)

        >>> point_a, point_b = cylinder.intersect_line(Line([0, 0, 0], [1, 0, 1]), infinite=False)

        >>> point_a
        Point([0., 0., 0.])
        >>> point_b
        Point([1., 0., 1.])

        """
        if line.dimension != 3:
            raise ValueError("The line must be 3D.")

        if infinite:
            return _intersect_line_with_infinite_cylinder(self, line, n_digits)

        return _intersect_line_with_finite_cylinder(self, line, n_digits)

    def to_mesh(self, n_along_axis: int = 100, n_angles: int = 30) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Return coordinate matrices for the 3D surface of the cylinder.

        Parameters
        ----------
        n_along_axis : int
            Number of intervals along the axis of the cylinder.
        n_angles : int
            Number of angles distributed around the circle.

        Returns
        -------
        X, Y, Z: (n_angles, n_angles) ndarray
            Coordinate matrices.

        Examples
        --------
        >>> from skspatial.objects import Cylinder

        >>> X, Y, Z = Cylinder([0, 0, 0], [0, 0, 1], 1).to_mesh(2, 4)

        >>> X.round(3)
        array([[-1. , -1. ],
               [ 0.5,  0.5],
               [ 0.5,  0.5],
               [-1. , -1. ]])

        >>> Y.round(3)
        array([[ 0.   ,  0.   ],
               [ 0.866,  0.866],
               [-0.866, -0.866],
               [-0.   , -0.   ]])

        >>> Z.round(3)
        array([[0., 1.],
               [0., 1.],
               [0., 1.],
               [0., 1.]])

        """
        # Unit vector along the cylinder axis.
        v_axis = self.vector.unit()

        # Arbitrary unit vector in a direction other than the axis.
        # This is used to get a vector perpendicular to the axis.
        v_different_direction = v_axis.different_direction()

        # Two unit vectors that are mutually perpendicular
        # and perpendicular to the cylinder axis.
        # These are used to define the points on the cylinder surface.
        u_1 = v_axis.cross(v_different_direction).unit()
        u_2 = v_axis.cross(u_1).unit()

        # The cylinder surface ranges over t from 0 to length of axis,
        # and over theta from 0 to 2 * pi.
        t = np.linspace(0, self.length(), n_along_axis)
        theta = np.linspace(0, 2 * np.pi, n_angles)

        # use meshgrid to make 2d arrays
        t, theta = np.meshgrid(t, theta)

        X, Y, Z = [
            self.point[i] + v_axis[i] * t + self.radius * np.sin(theta) * u_1[i] + self.radius * np.cos(theta) * u_2[i]
            for i in range(3)
        ]

        return X, Y, Z

    def plot_3d(self, ax_3d: Axes3D, n_along_axis: int = 100, n_angles: int = 30, **kwargs) -> None:
        """
        Plot a 3D cylinder.

        Parameters
        ----------
        ax_3d : Axes3D
            Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
        n_along_axis : int
            Number of intervals along the axis of the cylinder.
        n_angles : int
            Number of angles distributed around the circle.
        kwargs : dict, optional
            Additional keywords passed to :meth:`~mpl_toolkits.mplot3d.axes3d.Axes3D.plot_surface`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from mpl_toolkits.mplot3d import Axes3D

            >>> from skspatial.objects import Cylinder

            >>> fig = plt.figure()
            >>> ax = fig.add_subplot(111, projection='3d')

            >>> cylinder = Cylinder([5, 3, 1], [1, 0, 1], 2)

            >>> cylinder.plot_3d(ax, alpha=0.2)
            >>> cylinder.point.plot_3d(ax, s=100)

        """
        X, Y, Z = self.to_mesh(n_along_axis, n_angles)

        ax_3d.plot_surface(X, Y, Z, **kwargs)


def _between_cap_planes(cylinder: Cylinder, point: array_like) -> bool:
    """Check if a point lies between the cylinder cap planes."""
    plane_base = Plane(cylinder.point, cylinder.vector)
    distance_point_signed = plane_base.distance_point_signed(point)

    return 0 <= distance_point_signed <= distance_point_signed <= cylinder.length()


def _intersect_line_with_infinite_cylinder(
    cylinder: Cylinder,
    line: Line,
    n_digits: Optional[int],
) -> Tuple[Point, Point]:

    p_c = cylinder.point
    v_c = cylinder.vector.unit()
    r = cylinder.radius

    p_l = line.point
    v_l = line.vector.unit()

    delta_p = Vector.from_points(p_c, p_l)

    a = (v_l - v_l.dot(v_c) * v_c).norm() ** 2
    b = 2 * (v_l - v_l.dot(v_c) * v_c).dot(delta_p - delta_p.dot(v_c) * v_c)
    c = (delta_p - delta_p.dot(v_c) * v_c).norm() ** 2 - r**2

    try:
        X = _solve_quadratic(a, b, c, n_digits=n_digits)
    except ValueError:
        raise ValueError("The line does not intersect the cylinder.")

    point_a, point_b = p_l + X.reshape(-1, 1) * v_l

    return point_a, point_b


def _intersect_line_with_caps(cylinder: Cylinder, line: Line) -> Tuple[Optional[Point], Optional[Point]]:
    """Find the intersection points of the line with the cylinder caps."""

    def _intersect_cap(plane_cap: Plane) -> Optional[Point]:

        try:
            point_intersection = plane_cap.intersect_line(line)
        except ValueError:
            return None

        return point_intersection if point_intersection.distance_point(plane_cap.point) <= cylinder.radius else None

    # The planes containing the circular caps of the cylinder.
    plane_base = Plane(point=cylinder.point, normal=cylinder.vector)
    plane_top = Plane(point=cylinder.point + cylinder.vector, normal=cylinder.vector)

    point_base = _intersect_cap(plane_base)
    point_top = _intersect_cap(plane_top)

    return point_base, point_top


def _intersect_line_with_finite_cylinder(
    cylinder: Cylinder,
    line: Line,
    n_digits: Optional[int],
) -> Tuple[Point, Point]:
    """Find intersection points of a line with a cylinder, including the cylinder caps."""
    point_base, point_top = _intersect_line_with_caps(cylinder, line)

    if point_base is not None and point_top is not None:
        return point_base, point_top

    point_a, point_b = _intersect_line_with_infinite_cylinder(cylinder, line, n_digits)

    if not _between_cap_planes(cylinder, point_a):
        point_a = point_base if point_base is not None else point_top

    if not _between_cap_planes(cylinder, point_b):
        point_b = point_base if point_base is not None else point_top

    if point_a is None or point_b is None:
        raise ValueError("The line does not intersect the cylinder.")

    return point_a, point_b



import math

import numpy as np
import pytest

from skspatial.measurement import area_signed
from skspatial.measurement import area_triangle
from skspatial.measurement import volume_tetrahedron
from skspatial.objects import Points


@pytest.mark.parametrize(
    ("array_a", "array_b", "array_c", "area_expected"),
    [
        ([0, 0], [1, 0], [0, 1], 0.5),
        ([0, 0], [1, 1], [2, 0], 1),
        ([0, 0], [1, 10], [2, 0], 10),
        ([0, 0], [1, 0], [2, 0], 0),
        ([0, 0], [-5, -2], [5, 2], 0),
        ([1, 0, 0], [0, 1, 0], [0, 0, 1], math.sin(np.pi / 3)),
        ([2, 0, 0], [0, 2, 0], [0, 0, 2], 4 * math.sin(np.pi / 3)),
    ],
)
def test_area_triangle(array_a, array_b, array_c, area_expected):

    area = area_triangle(array_a, array_b, array_c)
    assert math.isclose(area, area_expected)


@pytest.mark.parametrize(
    ("array_a", "array_b", "array_c", "array_d", "volume_expected"),
    [
        ([0, 0], [2, 0], [1, 1], [10, -7], 0),
        ([0, 0, 0], [2, 0, 0], [1, 1, 0], [0, 0, 1], 1 / 3),
        ([0, 0, 0], [2, 0, 0], [1, 1, 0], [0, 0, -1], 1 / 3),
        ([0, 0, 0], [2, 0, 0], [1, 1, 0], [0, 0, 2], 2 / 3),
        ([0, 0, 0], [2, 0, 0], [1, 1, 0], [0, 0, 3], 1),
        ([0, 0, 0], [2, 0, 0], [1, 1, 0], [-56, 10, 3], 1),
        ([0, 1, 1], [0, 1, 5], [0, -5, 7], [0, 5, 2], 0),
    ],
)
def test_volume_tetrahedron(array_a, array_b, array_c, array_d, volume_expected):

    volume = volume_tetrahedron(array_a, array_b, array_c, array_d)
    assert math.isclose(volume, volume_expected)


@pytest.mark.parametrize(
    ("points", "area_expected"),
    [
        # Counter-clockwise triangle
        ([[2, 2], [6, 2], [4, 5]], 6),
        # Clockwise triangle
        ([[1, 3], [-4, 3], [-3, 4]], -2.5),
        # Counter-clockwise square
        ([[-1, 2], [2, 5], [-1, 8], [-4, 5]], 18),
        # Clockwise irregular convex pentagon
        ([[-2, 2], [-5, 2], [-8, 5], [-4, 8], [-1, 5]], -25.5),
        # Counter-clockwise irregular convex hexagon
        ([[3, -2], [6, -3], [10, -1], [8, 4], [4, 3], [1, 1]], 39.5),
        # Clockwise non-convex polygon
        ([[5, -2], [1, -1], [0, 4], [6, 6], [3, 3]], -22),
        # Self-overlapping polygon
        ([[-4, 4], [-4, 1], [2, 4], [2, 1]], 0),
    ],
)
def test_area_signed(points, area_expected):

    points = Points(points)
    area = area_signed(points)

    assert area == area_expected


@pytest.mark.parametrize(
    ("points", "message_expected"),
    [
        ([[1, 0, 0], [-1, 0, 0], [0, 1, 0]], "The points must be 2D."),
        ([[2, 0], [-2, 0]], "There must be at least 3 points."),
    ],
)
def test_area_signed_failure(points, message_expected):

    with pytest.raises(ValueError, match=message_expected):
        area_signed(points)



"""Private functions for some spatial computations."""
from __future__ import annotations

import math
from functools import wraps
from typing import Any
from typing import Callable
from typing import Optional

import numpy as np

from skspatial.typing import array_like


def _contains_point(obj: Any, point: array_like, **kwargs: float) -> bool:
    """
    Check if the object contains a point.

    Returns True if the distance from the point to the object is close to zero.

    Parameters
    ----------
    obj: Object
        Spatial object (e.g. Line).
    point : array_like
        Input point.
    kwargs : dict, optional
        Additional keywords passed to :func:`math.isclose`.

    Returns
    -------
    bool
        True if the spatial object contains the input point.

    Notes
    -----
    Setting an absolute tolerance is useful when comparing a value to zero.

    """
    distance = obj.distance_point(point)

    return math.isclose(distance, 0, **kwargs)


def _sum_squares(obj: Any, points: array_like) -> np.float64:
    """Return the sum of squared distances from points to a spatial object."""
    distances_squared = np.apply_along_axis(obj.distance_point, 1, points) ** 2

    return distances_squared.sum()


def _mesh_to_points(X: array_like, Y: array_like, Z: array_like) -> np.ndarray:
    """Convert a mesh into an (N, 3) array of N points."""
    return np.vstack([*map(np.ravel, [X, Y, Z])]).T


def np_float(func: Callable) -> Callable[..., np.float64]:
    """
    Cast the output type as np.float64.

    Outputs with type np.float64 have a useful round() method.

    """
    # wraps() is needed so that sphinx generates
    # the docstring of functions with this decorator.
    @wraps(func)
    def wrapper(*args):
        return np.float64(func(*args))

    return wrapper


def _solve_quadratic(a: float, b: float, c: float, n_digits: Optional[int] = None) -> np.ndarray:
    """
    Solve a quadratic equation.

    The equation has the form

    .. math:: ax^2 + bx + c = 0

    Parameters
    ----------
    a, b, c : float
        Coefficients of the quadratic equation.
    n_digits : int, optional
        Additional keyword passed to :func:`round` (default None).

    Returns
    -------
    np.ndarray
        Array containing the two solutions to the quadratic.

    Raises
    ------
    ValueError
        If the coefficient `a` is zero.
        If the discriminant is negative.

    Examples
    --------
    >>> from skspatial._functions import _solve_quadratic

    >>> _solve_quadratic(-1, 1, 1).round(3)
    array([ 1.618, -0.618])

    >>> _solve_quadratic(0, 1, 1)
    Traceback (most recent call last):
    ...
    ValueError: The coefficient `a` must be non-zero.

    >>> _solve_quadratic(1, 1, 1)
    Traceback (most recent call last):
    ...
    ValueError: The discriminant must not be negative.

    """
    if n_digits:
        a = round(a, n_digits)
        b = round(b, n_digits)
        c = round(c, n_digits)

    if a == 0:
        raise ValueError("The coefficient `a` must be non-zero.")

    discriminant = b**2 - 4 * a * c

    if discriminant < 0:
        raise ValueError("The discriminant must not be negative.")

    pm = np.array([-1, 1])  # Array to compute minus/plus.

    X = (-b + pm * math.sqrt(discriminant)) / (2 * a)

    return X


_allclose = np.vectorize(math.isclose)



"""Test functionality of objects based on a single 1D NumPy array (Point and Vector)."""
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from skspatial.objects import Point
from skspatial.objects import Vector


@pytest.mark.parametrize("class_spatial", [Point, Vector])
@pytest.mark.parametrize("array", [[1, 0], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 3, 4, 5]])
def test_equality(class_spatial, array):

    assert_array_equal(array, class_spatial(array))


@pytest.mark.parametrize("class_spatial", [Point, Vector])
@pytest.mark.parametrize(
    "array",
    [
        [[0]],
        [[0], [0]],
        [[0, 0], [0, 0]],
        [[1, 2, 3]],
    ],
)
def test_failure(class_spatial, array):

    message_expected = "The array must be 1D."

    with pytest.raises(ValueError, match=message_expected):
        class_spatial(array)


@pytest.mark.parametrize("class_spatial", [Point, Vector])
@pytest.mark.parametrize(
    ("array", "dim", "array_expected"),
    [
        ([0, 0], 2, [0, 0]),
        ([0, 0], 3, [0, 0, 0]),
        ([0, 0], 5, [0, 0, 0, 0, 0]),
        ([6, 3, 7], 4, [6, 3, 7, 0]),
    ],
)
def test_set_dimension(class_spatial, array, dim, array_expected):

    object_spatial = class_spatial(array).set_dimension(dim)
    assert object_spatial.is_close(array_expected)


@pytest.mark.parametrize("class_spatial", [Point, Vector])
@pytest.mark.parametrize(
    ("array", "dimension"),
    [
        (np.zeros(3), 2),
        (np.zeros(2), 1),
        (np.zeros(1), 0),
    ],
)
def test_dimension_failure(class_spatial, array, dimension):

    message_expected = "The desired dimension cannot be less than the current dimension."

    object_spatial = class_spatial(array)

    with pytest.raises(ValueError, match=message_expected):
        object_spatial.set_dimension(dimension)


@pytest.mark.parametrize("class_spatial", [Point, Vector])
def test_dimension_of_slice(class_spatial):

    object_spatial = class_spatial([0, 0, 0])

    assert object_spatial.dimension == 3
    assert object_spatial[:3].dimension == 3
    assert object_spatial[:2].dimension == 2
    assert object_spatial[:1].dimension == 1



"""Module for the Point class."""
import numpy as np
from matplotlib.axes import Axes
from mpl_toolkits.mplot3d import Axes3D

from skspatial.objects._base_array import _BaseArray1D
from skspatial.objects.vector import Vector
from skspatial.plotting import _scatter_2d
from skspatial.plotting import _scatter_3d
from skspatial.typing import array_like


class Point(_BaseArray1D):
    """
    A point in space implemented as a 1D array.

    The array is a subclass of :class:`numpy.ndarray`.

    Parameters
    ----------
    array : array_like
        Input array.

    Attributes
    ----------
    dimension : int
        Dimension of the point.

    Raises
    ------
    ValueError
        If the array is empty, the values are not finite,
        or the dimension is not one.

    Examples
    --------
    >>> from skspatial.objects import Point

    >>> point = Point([1, 2, 3])

    >>> point.dimension
    3

    The object inherits methods from :class:`numpy.ndarray`.

    >>> point.mean()
    array(2.)

    >>> Point([])
    Traceback (most recent call last):
    ...
    ValueError: The array must not be empty.

    >>> import numpy as np

    >>> Point([1, 2, np.nan])
    Traceback (most recent call last):
    ...
    ValueError: The values must all be finite.

    >>> Point([[1, 2], [3, 4]])
    Traceback (most recent call last):
    ...
    ValueError: The array must be 1D.

    """

    def distance_point(self, other: array_like) -> np.float64:
        """
        Return the distance to another point.

        Parameters
        ----------
        other : array_like
            Other point.

        Returns
        -------
        np.float64
            Distance between the points.

        Examples
        --------
        >>> from skspatial.objects import Point

        >>> point = Point([1, 2])
        >>> point.distance_point([1, 2])
        0.0

        >>> point.distance_point([-1, 2])
        2.0

        >>> Point([1, 2, 0]).distance_point([1, 2, 3])
        3.0

        """
        vector = Vector.from_points(self, other)

        return vector.norm()

    def plot_2d(self, ax_2d: Axes, **kwargs) -> None:
        """
        Plot the point on a 2D scatter plot.

        Parameters
        ----------
        ax_2d : Axes
            Instance of :class:`~matplotlib.axes.Axes`.
        kwargs : dict, optional
            Additional keywords passed to :meth:`~matplotlib.axes.Axes.scatter`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from skspatial.objects import Point

            >>> _, ax = plt.subplots()

            >>> Point([1, 2]).plot_2d(ax, c='k', s=100)

        """
        _scatter_2d(ax_2d, self.reshape(1, -1), **kwargs)

    def plot_3d(self, ax_3d: Axes3D, **kwargs) -> None:
        """
        Plot the point on a 3D scatter plot.

        Parameters
        ----------
        ax_3d : Axes3D
            Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
        kwargs : dict, optional
            Additional keywords passed to :meth:`~mpl_toolkits.mplot3d.axes3d.Axes3D.scatter`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from mpl_toolkits.mplot3d import Axes3D

            >>> from skspatial.objects import Point

            >>> fig = plt.figure()
            >>> ax = fig.add_subplot(111, projection='3d')

            >>> Point([1, 2, 3]).plot_3d(ax, c='k', s=100)

        """
        _scatter_3d(ax_3d, self.reshape(1, -1), **kwargs)



import numpy as np
import pytest

from skspatial.objects import Points


@pytest.mark.parametrize(
    ("array", "dimension"),
    [
        (np.zeros((3, 1)), 0),
        (np.zeros((3, 2)), 1),
        (np.zeros((3, 3)), 2),
    ],
)
def test_dimension_failure(array, dimension):

    message_expected = "The desired dimension cannot be less than the current dimension."

    points = Points(array)

    with pytest.raises(ValueError, match=message_expected):
        points.set_dimension(dimension)



"""Module for the Vector class."""
from __future__ import annotations

import math
from typing import cast

import numpy as np
from matplotlib.axes import Axes
from mpl_toolkits.mplot3d import Axes3D

from skspatial._functions import np_float
from skspatial.objects._base_array import _BaseArray1D
from skspatial.plotting import _connect_points_3d
from skspatial.typing import array_like


class Vector(_BaseArray1D):
    """
    A vector implemented as a 1D array.

    The array is a subclass of :class:`numpy.ndarray`.

    Parameters
    ----------
    array : array_like
        Input array.

    Attributes
    ----------
    dimension : int
        Dimension of the vector.

    Raises
    ------
    ValueError
        If the array is empty, the values are not finite,
        or the dimension is not one.

    Examples
    --------
    >>> from skspatial.objects import Vector

    >>> vector = Vector([1, 2, 3])

    >>> vector.dimension
    3

    The object inherits methods from :class:`numpy.ndarray`.

    >>> vector.mean()
    array(2.)

    >>> Vector([])
    Traceback (most recent call last):
    ...
    ValueError: The array must not be empty.

    >>> import numpy as np

    >>> Vector([1, 2, np.nan])
    Traceback (most recent call last):
    ...
    ValueError: The values must all be finite.

    >>> Vector([[1, 2], [3, 4]])
    Traceback (most recent call last):
    ...
    ValueError: The array must be 1D.

    """

    @classmethod
    def from_points(cls, point_a: array_like, point_b: array_like) -> Vector:
        """
        Instantiate a vector from point A to point B.

        Parameters
        ----------
        point_a, point_b : array_like
            Points defining the vector.

        Returns
        -------
        Vector
            Vector from point A to point B.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector.from_points([0, 0], [1, 0])
        Vector([1, 0])

        >>> Vector.from_points([5, 2], [-2, 8])
        Vector([-7,  6])

        >>> Vector.from_points([3, 1, 1], [7, 7, 0])
        Vector([ 4,  6, -1])

        """
        array_vector_ab = cast(np.ndarray, np.subtract(point_b, point_a))

        return cls(array_vector_ab)

    def norm(self, **kwargs) -> np.float64:
        """
        Return the norm of the vector.

        Parameters
        ----------
        kwargs : dict, optional
            Additional keywords passed to :func:`numpy.linalg.norm`.

        Returns
        -------
        np.float64
            Norm of the vector.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> vector = Vector([1, 2, 3])

        >>> vector.norm().round(3)
        3.742

        >>> vector.norm(ord=1)
        6.0

        >>> vector.norm(ord=0)
        3.0

        """
        return np.linalg.norm(self, **kwargs)

    def unit(self) -> Vector:
        """
        Return the unit vector in the same direction as the vector.

        A unit vector is a vector with a magnitude of one.

        Returns
        -------
        Vector
            Unit vector.

        Raises
        ------
        ValueError
            If the magnitude of the vector is zero.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector([1, 0]).unit()
        Vector([1., 0.])

        >>> Vector([-20, 0]).unit()
        Vector([-1.,  0.])

        >>> Vector([1, 1]).unit().round(3)
        Vector([0.707, 0.707])

        >>> Vector([1, 1, 1]).unit().round(3)
        Vector([0.577, 0.577, 0.577])

        >>> Vector([0, 0]).unit()
        Traceback (most recent call last):
        ...
        ValueError: The magnitude must not be zero.

        """
        magnitude = self.norm()

        if magnitude == 0:
            raise ValueError("The magnitude must not be zero.")

        return self / magnitude

    def is_zero(self, **kwargs: float) -> bool:
        """
        Check if the vector is the zero vector.

        The zero vector in n dimensions is the vector containing n zeros.

        Parameters
        ----------
        kwargs : dict, optional
            Additional keywords passed to :func:`math.isclose`.

        Returns
        -------
        bool
            True if vector is the zero vector; false otherwise.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector([0, 0]).is_zero()
        True
        >>> Vector([1, 0]).is_zero()
        False

        >>> Vector([0, 0, 1e-4]).is_zero()
        False
        >>> Vector([0, 0, 1e-4]).is_zero(abs_tol=1e-3)
        True

        """
        return math.isclose(self.dot(self), 0, **kwargs)

    def cross(self, other: array_like) -> Vector:
        """
        Compute the cross product with another vector.

        Parameters
        ----------
        other : array_like
             Other vector.

        Returns
        -------
        Vector
            3D vector perpendicular to both inputs.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector([1, 0]).cross([0, 1])
        Vector([0, 0, 1])

        >>> Vector([2, 5]).cross([1, 1])
        Vector([ 0,  0, -3])

        >>> Vector([1, 0]).cross([0, 1])
        Vector([0, 0, 1])

        >>> Vector([1, 1, 1]).cross([0, 1, 0])
        Vector([-1,  0,  1])

        """
        # Convert to 3D vectors so that cross product is also 3D.
        vector_a = self.set_dimension(3)
        vector_b = Vector(other).set_dimension(3)

        return Vector(np.cross(vector_a, vector_b))

    def cosine_similarity(self, other: array_like) -> np.float64:
        """
        Return the cosine similarity of the vector with another.

        This is the cosine of the angle between the vectors.

        Parameters
        ----------
        other : array_like
            Other vector.

        Returns
        -------
        np.float64
            Cosine similarity.

        Raises
        ------
        ValueError
            If either vector has a magnitude of zero.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector([1, 0]).cosine_similarity([0, 1])
        0.0

        >>> Vector([30, 0]).cosine_similarity([0, 20])
        0.0

        >>> Vector([1, 0]).cosine_similarity([-1, 0])
        -1.0

        >>> Vector([1, 0]).cosine_similarity([1, 1]).round(3)
        0.707

        >>> Vector([0, 0]).cosine_similarity([1, 1])
        Traceback (most recent call last):
        ...
        ValueError: The vectors must have non-zero magnitudes.

        """
        denom = self.norm() * Vector(other).norm()

        if denom == 0:
            raise ValueError("The vectors must have non-zero magnitudes.")

        cos_theta = self.dot(other) / denom

        # Ensure that the output is in the range [-1, 1],
        # so that the angle theta is defined.
        clipped = np.clip(cos_theta, -1, 1)

        return np.float64(clipped)

    @np_float
    def angle_between(self, other: array_like) -> float:
        """
        Return the angle in radians between the vector and another.

        Parameters
        ----------
        other : array_like
            Other vector.

        Returns
        -------
        np.float64
            Angle between vectors in radians.

        Examples
        --------
        >>> import numpy as np
        >>> from skspatial.objects import Vector

        >>> Vector([1, 0]).angle_between([1, 0])
        0.0

        >>> Vector([1, 1, 1]).angle_between([1, 1, 1])
        0.0

        >>> angle = Vector([1, 0]).angle_between([1, 1])
        >>> np.degrees(angle).round()
        45.0

        >>> angle = Vector([1, 0]).angle_between([-2, 0])
        >>> np.degrees(angle).round()
        180.0

        """
        cos_theta = self.cosine_similarity(other)

        return math.acos(cos_theta)

    @np_float
    def angle_signed(self, other: array_like) -> float:
        """
        Return the signed angle in radians between the vector and another.

        The vectors must be 2D.

        Parameters
        ----------
        other : array_like
            Other vector.

        Returns
        -------
        np.float64
            Signed angle between vectors in radians.

        Raises
        ------
        ValueError
            If the vectors are not 2D.

        Examples
        --------
        >>> import numpy as np
        >>> from skspatial.objects import Vector

        >>> Vector([1, 0]).angle_signed([1, 0])
        0.0

        >>> np.degrees(Vector([1, 0]).angle_signed([0, 1]))
        90.0

        >>> np.degrees(Vector([1, 0]).angle_signed([0, -1]))
        -90.0

        >>> Vector([1, 0, 0]).angle_signed([0, -1, 0])
        Traceback (most recent call last):
        ...
        ValueError: The vectors must be 2D.

        """
        if not (self.dimension == 2 and Vector(other).dimension == 2):
            raise ValueError("The vectors must be 2D.")

        dot = self.dot(other)
        det = np.linalg.det([self, other])

        return math.atan2(det, dot)

    def is_perpendicular(self, other: array_like, **kwargs: float) -> bool:
        r"""
        Check if the vector is perpendicular to another.

        Two vectors :math:`u` and :math:`v` are perpendicular if

        .. math::
            u \cdot v = 0

        Parameters
        ----------
        other : array_like
            Other vector.
        kwargs : dict, optional
            Additional keywords passed to :func:`math.isclose`.

        Returns
        -------
        bool
            True if the vector is perpendicular; false otherwise.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector([0, 1]).is_perpendicular([1, 0])
        True

        >>> Vector([-1, 5]).is_perpendicular([3, 4])
        False

        >>> Vector([2, 0, 0]).is_perpendicular([0, 0, 2])
        True

        The zero vector is perpendicular to all vectors.

        >>> Vector([0, 0, 0]).is_perpendicular([1, 2, 3])
        True

        """
        return math.isclose(self.dot(other), 0, **kwargs)

    def is_parallel(self, other: array_like, **kwargs: float) -> bool:
        r"""
        Check if the vector is parallel to another.

        Two nonzero vectors :math:`u` and :math:`v` are parallel if

        .. math::
            \texttt{abs}(\texttt{cosine_similarity}(u, v)) = 1

        The zero vector is parallel to all vectors.

        Parameters
        ----------
        other : array_like
            Other vector.
        kwargs : dict, optional
            Additional keywords passed to :func:`math.isclose`.

        Returns
        -------
        bool
            True if the vector is parallel; false otherwise.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector([0, 1]).is_parallel([1, 0])
        False

        >>> Vector([1, 1]).is_parallel([1, 1])
        True

        >>> Vector([-1, 5]).is_parallel([2, -10])
        True

        >>> Vector([1, 2, 3]).is_parallel([3, 6, 9])
        True

        >>> Vector([1, 2, 3, 4]).is_parallel([-2, -4, -6, -8])
        True

        The zero vector is parallel to all vectors.

        >>> Vector([1, 2, 3]).is_parallel([0, 0, 0])
        True

        """
        if self.is_zero(**kwargs) or Vector(other).is_zero(**kwargs):
            # The zero vector is perpendicular to all vectors.
            return True

        similarity = self.cosine_similarity(other)

        return math.isclose(abs(similarity), 1, **kwargs)

    def side_vector(self, other: array_like) -> int:
        """
        Find the side of the vector where another vector is directed.

        Both vectors must be 2D.

        Parameters
        ----------
        other : array_like
            Other 2D vector.

        Returns
        -------
        int
            1 if the other vector is to the right.
            0 if the other is parallel.
            -1 if the other is to the left.

        Raises
        ------
        ValueError
            If the vectors are not 2D.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> vector_target = Vector([0, 1])

        The vector is parallel to the target vector.

        >>> vector_target.side_vector([0, 2])
        0
        >>> vector_target.side_vector([0, -5])
        0

        The vector is to the right of the target vector.

        >>> vector_target.side_vector([1, 1])
        1
        >>> vector_target.side_vector([1, -10])
        1

        The vector is to the left of the target vector.

        >>> vector_target.side_vector([-3, 4])
        -1

        The vectors are not 2D.

        >>> Vector([1]).side_vector([2])
        Traceback (most recent call last):
        ...
        ValueError: The vectors must be 2D.

        >>> Vector([1, 0, 0]).side_vector([1, 2, 3])
        Traceback (most recent call last):
        ...
        ValueError: The vectors must be 2D.

        """
        if self.dimension != 2 or Vector(other).dimension != 2:
            raise ValueError("The vectors must be 2D.")

        value_cross = np.cross(other, self)

        return int(np.sign(value_cross))

    def scalar_projection(self, other: array_like) -> np.float64:
        """
        Return the scalar projection of an other vector onto the vector.

        Parameters
        ----------
        other : array_like
            Other vector.

        Returns
        -------
        np.float64
            Scalar projection.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector([0, 1]).scalar_projection([2, 1])
        1.0

        >>> Vector([-1, -1]).scalar_projection([1, 0]).round(3)
        -0.707

        >>> Vector([0, 100]).scalar_projection([9, 5])
        5.0

        >>> Vector([5, 0]).scalar_projection([-10, 3])
        -10.0

        """
        result = self.unit().dot(other)

        return np.float64(result)

    def project_vector(self, other: array_like) -> Vector:
        """
        Project an other vector onto the vector.

        Parameters
        ----------
        other : array_like
            Other vector.

        Returns
        -------
        Vector
            Vector projection.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector([0, 1]).project_vector([2, 1])
        Vector([0., 1.])

        >>> Vector([0, 100]).project_vector([2, 1])
        Vector([0., 1.])

        >>> Vector([0, 1]).project_vector([9, 5])
        Vector([0., 5.])

        >>> Vector([0, 100]).project_vector([9, 5])
        Vector([0., 5.])

        """
        return self.dot(other) / self.dot(self) * self

    def different_direction(self, **kwargs: float) -> Vector:
        """
        Return a simple vector that is in a different direction.

        This is useful for finding a vector perpendicular to the original,
        by taking the cross product of the original with the one in a different direction.

        Parameters
        ----------
        kwargs : dict, optional
            Additional keywords passed to :meth:`Vector.is_zero` and :meth:`Vector.is_parallel`.
            :meth:`Vector.is_zero` is used to ensure the input vector is not the zero vector,
            and :meth:`Vector.is_parallel` is used to ensure the new vector is not parallel to the input.

        Returns
        -------
        Vector
            A unit vector in a different direction from the original.

        Raises
        ------
        ValueError
            If the vector is the zero vector.

        Examples
        --------
        >>> from skspatial.objects import Vector

        >>> Vector([1]).different_direction()
        Vector([-1])
        >>> Vector([100]).different_direction()
        Vector([-1])
        >>> Vector([-100]).different_direction()
        Vector([1])
        >>> Vector([1, 0]).different_direction()
        Vector([0., 1.])
        >>> Vector([1, 1]).different_direction()
        Vector([1., 0.])
        >>> Vector([1, 1, 1, 1]).different_direction()
        Vector([1., 0., 0., 0.])

        """
        if self.is_zero(**kwargs):
            raise ValueError("The vector must not be the zero vector.")

        if self.dimension == 1:
            return Vector([-np.sign(self[0])])

        vector_different_direction = Vector(np.zeros(self.dimension))
        vector_different_direction[0] = 1

        if self.is_parallel(vector_different_direction, **kwargs):
            vector_different_direction[0] = 0
            vector_different_direction[1] = 1

        return vector_different_direction

    def plot_2d(self, ax_2d: Axes, point: array_like = (0, 0), scalar: float = 1, **kwargs) -> None:
        """
        Plot a 2D vector.

        The vector is plotted as an arrow.

        Parameters
        ----------
        ax_2d : Axes
            Instance of :class:`~matplotlib.axes.Axes`.
        point : array_like, optional
            Position of the vector tail (default is origin).
        scalar : {int, float}, optional
            Value used to scale the vector (default 1).
        kwargs : dict, optional
            Additional keywords passed to :meth:`~matplotlib.axes.Axes.arrow`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from skspatial.objects import Vector

            >>> _, ax = plt.subplots()

            >>> Vector([1, 1]).plot_2d(ax, point=(-3, 5), scalar=2, head_width=0.5)

            >>> limits = ax.axis([-5, 5, 0, 10])

        """
        x, y = point
        dx, dy = scalar * self

        ax_2d.arrow(x, y, dx, dy, **kwargs)

    def plot_3d(self, ax_3d: Axes3D, point: array_like = (0, 0, 0), scalar: float = 1, **kwargs) -> None:
        """
        Plot a 3D vector.

        The vector is plotted by connecting two 3D points
        (the head and tail of the vector).

        Parameters
        ----------
        ax_3d : Axes3D
            Instance of :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.
        point : array_like, optional
            Position of the vector tail (default is origin).
        scalar : {int, float}, optional
            Value used to scale the vector (default 1).
        kwargs : dict, optional
            Additional keywords passed to :meth:`~mpl_toolkits.mplot3d.axes3d.Axes3D.plot`.

        Examples
        --------
        .. plot::
            :include-source:

            >>> import matplotlib.pyplot as plt
            >>> from mpl_toolkits.mplot3d import Axes3D

            >>> from skspatial.objects import Vector

            >>> fig = plt.figure()
            >>> ax = fig.add_subplot(111, projection='3d')

            >>> Vector([-1, 1, 1]).plot_3d(ax, point=(1, 2, 3), c='r')

        """
        point_2 = np.array(point) + scalar * self

        _connect_points_3d(ax_3d, point, point_2, **kwargs)



"""Top-level package for scikit-spatial."""

__author__ = "Andrew Hynes"
__email__ = "andrewjhynes@gmail.com"

try:
    import importlib.metadata as importlib_metadata  # type: ignore

except ModuleNotFoundError:
    import importlib_metadata  # type: ignore

__version__ = importlib_metadata.version("scikit-spatial")



"""Spatial transformations."""
from typing import cast
from typing import Sequence

import numpy as np

from skspatial.typing import array_like


def transform_coordinates(points: array_like, point_origin: array_like, vectors_basis: Sequence) -> np.ndarray:
    """
    Transform points into a new coordinate system.

    Parameters
    ----------
    points : (N, D) array_like
        Array of N points with dimension D.
    point_origin : (D,) array_like
        Origin of the new coordinate system.
        Array for one point with dimension D.
    vectors_basis : sequence
        Basis vectors of the new coordinate system.
        Sequence of N_bases vectors.
        Each vector is an array_like with D elements.

    Returns
    -------
    ndarray
        Coordinates in the new coordinate system.
        (N, N_bases) array.

    Examples
    --------
    >>> from skspatial.transformation import transform_coordinates

    >>> points = [[1, 2], [3, 4], [5, 6]]
    >>> vectors_basis = [[1, 0], [1, 1]]

    >>> transform_coordinates(points, [0, 0], vectors_basis)
    array([[ 1,  3],
           [ 3,  7],
           [ 5, 11]])

    >>> points = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    >>> vectors_basis = [[1, 0, 0], [-1, 1, 0]]

    >>> transform_coordinates(points, [0, 0, 0], vectors_basis)
    array([[1, 1],
           [4, 1],
           [7, 1]])

    """
    vectors_to_points = np.subtract(points, point_origin)
    array_transformed = np.matmul(vectors_to_points, np.transpose(vectors_basis))

    return cast(np.ndarray, array_transformed)



"""Test functionality of objects based on a NumPy array (Point, Vector, and Points)."""
import numpy as np
import pytest

from skspatial.objects import Point
from skspatial.objects import Points
from skspatial.objects import Vector


@pytest.mark.parametrize("class_spatial", [Point, Vector, Points])
@pytest.mark.parametrize(
    "array",
    [
        [[0], [0, 0]],
        [[0, 0], [0, 0, 0]],
        [[0, 0, 0], [0, 0, 0], [0]],
    ],
)
def test_failure_from_different_lengths(class_spatial, array):

    message_expected = "The array must not contain sequences with different lengths."

    with pytest.raises(ValueError, match=message_expected):
        class_spatial(array)


@pytest.mark.parametrize("class_spatial", [Point, Vector, Points])
@pytest.mark.parametrize(
    "array",
    [
        [],
        [[]],
        [[], []],
    ],
)
def test_failure_from_empty_array(class_spatial, array):

    message_expected = "The array must not be empty."

    with pytest.raises(ValueError, match=message_expected):
        class_spatial(array)


@pytest.mark.parametrize("class_spatial", [Point, Vector, Points])
@pytest.mark.parametrize(
    "array",
    [
        [np.nan, 0],
        [1, 1, np.nan],
        [1, 1, np.inf],
        [[1, 1], [np.inf, 0]],
    ],
)
def test_failure_from_infinite_values(class_spatial, array):

    message_expected = "The values must all be finite."

    with pytest.raises(ValueError, match=message_expected):
        class_spatial(array)
